"""Module providing slideshow scheduler."""

import schedule
import subprocess

from kivy.clock import Clock
from kivy.logger import Logger, LOG_LEVELS
from schedule import ScheduleValueError

from repository import ConfigError, check_param, check_valid_required

from .controller import DISPLAY_MODE, PLAY_STATE


class Scheduler:
    """Slideshow scheduler.

    The slideswhow scheduler implements a simple, YAML/dict configurable
    scheduler for the display of slideshow during certain periods.
    The scheduler uses the python schedule package [1]. The run_pending method
    of the default scheduler is triggered by the Kivy Clock.
    Scheduling is event based. Events can be be used to turn the display on or
    off and select the slideshow to be displayed.

    [1] https://github.com/dbader/schedule
    """

    # Required and valid configuration parameters
    CONF_REQ_KEYS = {'time'}
    CONF_VALID_KEYS = {'display_mode', 'display_timeout', 'play_state', 'slideshow'} | CONF_REQ_KEYS

    def __init__(self, config, app):
        """ Initialize scheduler instance.

        :param config: Scheduler configuration from the corresponding
            configuration file section.
        :type config: dict
        :param app: Pyframe application
        :type app: pyframe.App
        :raises: ConfigError
        """
        self._app = app
        self._event = None

        # Build schedule from configuration
        Logger.info("Scheduler: Building schedule from configuration.")
        for event, event_config in config.items():

            try:
                # Check the configuration for valid and required parameters.
                check_valid_required(event_config, self.CONF_VALID_KEYS, self.CONF_REQ_KEYS)
                # Check parameter values.
                check_param('display_mode', event_config, required=False, options={ item.value for item in DISPLAY_MODE })
                check_param('play_state', event_config, required=False, options={ item.value for item in PLAY_STATE })
                check_param('slideshow', event_config, required=False, options=self._app.slideshows)
                check_param('display_timeout', event_config, required=False, is_int=True, gr=0)
                # Schedule event.
                schedule.every().day.at(event_config['time']).do(self.on_event, event, event_config)
                Logger.info(f"Scheduler: Event '{event}' scheduled at '{event_config['time']}'.")
            # Catch all configuration and schedule errors.
            except (ConfigError, TypeError, ScheduleValueError) as e:
                raise ConfigError(f"Error in the configuration of event '{event}'. {e}", event_config)

        # Turn display off
        self._app.display_off()
        # Set clock interval and callback function.
        self.run_pending(0)

    def on_event(self, event, config):
        """Handle scheduled events."""
        Logger.info(f"Scheduler: Event '{event}' fired.")
        # Set display mode if specified.
        display_mode = config.get('display_mode')
        if display_mode is not None:
            self._app.display_mode = display_mode
        # Set display timeout if specified.
        display_timeout = config.get('display_timeout')
        if display_timeout is not None:
            self._app.display_timeout = display_timeout
        # Set slideshow if specified, turn display on and start playing.
        slideshow = config.get('slideshow')
        if slideshow is not None:
            self._app.slideshow = slideshow
        # Set display state if specified.
        play_state = config.get('play_state')
        if play_state is not None:
            self._app.play_state = play_state

    def run_pending(self, dt):
        """Run pending schedule jobs."""
        schedule.run_pending()
        self._event = Clock.schedule_once(self.run_pending, schedule.idle_seconds())

    def stop(self):
        """Stop scheduler."""
        if self._event is not None:
            self._event.cancel()
            self._event = None



"""Module providing slideshow class."""

import tracemalloc
tracemalloc.start()

from repository import Index, RepositoryFile, IoError

from kivy.app import App
from kivy.clock import Clock
from kivy.core.window import Window
from kivy.logger import Logger
from kivy.uix.anchorlayout import AnchorLayout
from kivy.uix.widget import Widget

from repository import SORT_DIR, SORT_ORDER, check_param, check_valid_required

from .content import ErrorMessage, SlideshowImage, SlideshowVideo
from .controller import PLAY_STATE


class Slideshow(AnchorLayout):
    """Slideshow widget.

    The slideshow widget iterates through files in an index (repository.Index)
    and displays them by creating the corresponding widgets. Currently, images (SlideshowImage) and videos (SlideshowVideo) are supported.
    Slideshow behavior can be influenced by the slide show configuration and
    optional parameters, which are passed to the constructor.
    """

    # Required and valid configuration parameters
    CONF_REQ_KEYS = {'bg_color', 'enable_animation', 'label_content', 'label_duration', 'label_font_size', 'label_mode', 'label_padding', 'max_zoom', 'pause', 'resize', 'rotation'} | Index.CRIT_REQ_KEYS
    CONF_VALID_KEYS = {'always_excluded_tags', 'sound'} | CONF_REQ_KEYS | Index.CRIT_VALID_KEYS

    def __init__(self, name, index, config):
        """Initialize slideshow instance.

        :param name: name of slideshow.
        :type name: str
        :param index: index of files in active repositories.
        :type index: repository.Index
        :param config: slideshow configuration from configuration file section.
        :type config: dict
        :raises ConfigError:
        """
        AnchorLayout.__init__(self, anchor_x='center', anchor_y='center')
        # Basic intialization
        self._name = name
        self._index = index
        self._config = config
        self._play_state = PLAY_STATE.STOPPED
        self._next_event = None
        self._current_widget = None
        self._iterator = None

        self.cur_snapshot = None

        # Check the configuration for valid and required parameters.
        check_valid_required(config, self.CONF_VALID_KEYS, self.CONF_REQ_KEYS)
        # Convert from boolean to "on" (True) and "off" (False) if necessary.
        # This is a pecularity of the YAML 1.1 standard, which interprets "on"
        # and "off" as boolean values.
        if 'label_mode' in config:
            map = {True: "on", False: "off", "on": "on", "off": "off", "auto": "auto"}
            config['label_mode'] = map[config['label_mode']]
        # Check parameter values.
        check_param('pause', config, is_int=True, gr=0)
        check_param('rotation', config, is_int=True, options={0, 90, 180, 270})
        check_param('bg_color', config, is_color=True)
        check_param('label_content', config, options={"description", "full", "short"})
        check_param('label_duration', config, is_int=True, gr=0)
        check_param('label_font_size', config, gr=0, le=0.2)
        check_param('label_mode', config, options={"auto", "off", "on"})
        check_param('label_padding', config, gr=0, le=0.2)
        check_param('resize', config, options={"fit", "fill"})

        # Compile filter criteria for index iteration.
        # Extract all relevant parameters from the slideshow configuration.
        self._criteria = { key: config[key] for key in Index.CRIT_VALID_KEYS if key in config }

        # Map file types.
        if 'types' in config:
            check_param('types', config, recurse=True, options={"images", "videos"})
            value = config['types']
            # Convert to list if single value specified.
            if type(value) == str: value = [value]
            map = {'images': RepositoryFile.TYPE_IMAGE, 'videos': RepositoryFile.TYPE_VIDEO}
            self._criteria['types'] = [ map[type] for type in value ]

        # Map orientation.
        if 'orientation' in config:
            check_param('orientation', config, options={"landscape", "portrait"})
            value = config['orientation']
            map = {'landscape': RepositoryFile.ORIENTATION_LANDSCAPE, 'portrait': RepositoryFile.ORIENTATION_PORTRAIT}
            self._criteria['orientation'] = map[value]

        # Add always excluded tags.
        if 'always_excluded_tags' in config:
            check_param('always_excluded_tags', config, recurse=True, is_str=True)
            check_param('excluded_tags', config, required=False, recurse=True, is_str=True)
            value = config['always_excluded_tags']
            # Convert to list if single value specified.
            if type(value) == str: value = [value]
            # Add or append criterion.
            self._criteria['excluded_tags'] = self._criteria.get('excluded_tags', []) + value

        # Register event fired upon slideshow content changes.
        self.register_event_type('on_content_change')

    def _create_widget(self, file):
        """Create widget for display of the specified file.

        :param file: file to be displayed
        :type file: repository.File
        :rtype: Widget
        """
        if file.type == RepositoryFile.TYPE_IMAGE:
            widget = SlideshowImage(file, self._config)
        elif file.type == RepositoryFile.TYPE_VIDEO:
            widget = SlideshowVideo(file, self._config)
        else:
            widget = ErrorMessage(f"Type of file '{file.uuid}' is not supported.", self._config)
        return widget

    def _create_next_widget(self, previous=False):
        """Return widget for the next file in the slideshow.

        Alternatively, a widget for the previous file in the slideshow may be
        returend by setting the 'previous' flag to True.

        Catches any exception that occurs during file retrieval and skips the
        respective file. An empty widget is returned after 5 failed attempts.
        If the end of the iteration is reached, the iteration is restarted.

        :param previous: Set to True if widget for previous file shall be
            returned.
        :type previous: bool
        :return: next file in the slideshow.
        :rtype: Widget
        """
        # Return error message widget if the slideshow does not contain any
        # files and re-create index iterator.
        if self._iterator.length == 0:
            Logger.error("Slideshow: The slideshow does not contain any files.")
            # Re-create index iterator for a new chance. Possibly, the
            # background indexer has added new files in the meantime.
            self._iterator = self._index.iterator(**self._criteria)
            return ErrorMessage("The slideshow does not contain any files.", self._config)

        # Make up to 5 attempts to create the next content widget. Return an
        # error message widget otherwise.
        attempts = 0
        while True:
            try:
                # Attempt to retrieve next file in repository.
                if previous is False:
                    file = next(self._iterator)
                # Or attempt to retrieve previous file in repository if previous
                # flag is set.
                else:
                    file = self._iterator.previous()
                widget = self._create_widget(file)
                # Exit loop if no exception occurred.
                break
            # Create new iterator if end of slideshow has been reached and try
            # again.
            except StopIteration:
                Logger.info("Slideshow: End of slideshow reached. Restarting slideshow.")
                self._iterator = self._index.iterator(**self._criteria)
                # Make sure to return next and not previous file.
                previous = False
                continue
            # Log error if any other exception occurred and try again.
            except IoError as e:
                Logger.error(f"Slideshow: An I/O error occurred while retrieving the next file. {e}.")
                continue
            finally:
                attempts = attempts + 1
                # Restart slideshow and return error message widget if number of
                # failed attempts exceeds 5.
                if attempts >= 5:
                    Logger.error(f"Slideshow: Restarting slideshow after {attempts} failed attempts to retrieve the next file.")
                    # Re-create index iterator for a new chance. Possibly, the
                    # background indexer has updated the index in the meantime.
                    self._iterator = self._index.iterator(**self._criteria)
                    return ErrorMessage(f"Restarting slideshow after {attempts} failed attempts to retrieve the next file.", self._config)
        return widget

    @property
    def file_count(self):
        """Return number of files in the slideshow."""
        # Return initial count of current iterator if exists.
        if self._iterator is not None:
            return self._iterator.length
        else:
            # Create temporary iterator otherwise to determine length.
            return self._index.iterator(**self._criteria).length

    @property
    def current_file(self):
        """Return linked repository file for the current content widget.

        Note that the method may return None if no currrent file is available,
        e.g. if the slideshow is stopped.

        :return: linked repository file
        :rtype: repository.file
        """
        if self._current_widget is not None:
            return self._current_widget.file
        else: return None

    @property
    def name(self):
        """Return name of the slidshow.

        :return: name of slideshow
        :rtype: str
        """
        return self._name

    @property
    def play_state(self):
        """Return play state.

        See enumeration PLAY_STATE for possible values.

        :return: play state
        :rtype: str
        """
        return self._play_state

    def next(self, reschedule=True, previous=False):
        """Display next file in index."""
        # Skip if not playing or paused.
        if self._play_state == PLAY_STATE.STOPPED: return
        # Unschedule and re-schedule callback function.
        if reschedule and self._play_state == PLAY_STATE.PLAYING and self._next_event is not None:
            self._next_event.cancel()
            self._next_event = Clock.schedule_interval(self._clock_callback, self._config['pause'])
        # Remove current widget from layout.
        if self._current_widget is not None:
            self.remove_widget(self._current_widget)
        # Make widget from next file the current widget.
        self._current_widget = self._create_next_widget(previous)

        # Monitor memory usage. Code initially inserted to investigate
        # memory leakage problem observed with Radxa zero.
#        self.prev_snapshot = self.cur_snapshot
#        self.cur_snapshot = tracemalloc.take_snapshot()
#         if self.prev_snapshot:
#            stats = self.cur_snapshot.compare_to(self.prev_snapshot, 'lineno')
#        else:
#            stats = self.cur_snapshot.statistics('lineno')
#        Logger.debug("Top 10 memory consumption:")
#        for stat in stats[:10]:
#            Logger.debug(f"\t{stat}")

        # Stop playing content in current widget if slideshow is paused.
        if self._play_state == PLAY_STATE.PAUSED:
            self._current_widget.stop()
        # Add current widget to layout.
        self.add_widget(self._current_widget)
        # Fire event to indicate content change
        self.dispatch('on_content_change', self)

    def on_content_change(self, *largs):
        """Default handler for 'on_content_change' event."""
        # Do nothing.
        Logger.warn("Slideshow: Ignoring event 'on_content_change' since it should have been handled elsewhere.")

    def pause(self):
        """Pause playing slideshow."""
        # Skip if already paused or stopped.
        if self._play_state == PLAY_STATE.PAUSED or self._play_state == PLAY_STATE.STOPPED: return
        # Unschedule the callback function.
        if self._next_event is not None:
            self._next_event.cancel()
            self._next_event = None
        # Stop playing content in current widget.
        self._current_widget.stop()
        # Update state.
        self._play_state = PLAY_STATE.PAUSED

    def play(self):
        """Start playing slideshow."""
        # Skip if already playing.
        if self._play_state == PLAY_STATE.PLAYING: return
        # Create new selective index iterator with sorting/filter criteria from
        # the slideshow configuration if not paused.
        if self._play_state != PLAY_STATE.PAUSED:
            self._iterator = self._index.iterator(**self._criteria)
        # Remove current widget from layout.
        if self._current_widget is not None:
            self.remove_widget(self._current_widget)
        # Create current widget from first file and add to layout.
        self._current_widget = self._create_next_widget()
        self.add_widget(self._current_widget)
        # Schedule callback function to start playing slideshow.
        self._next_event = Clock.schedule_interval(self._clock_callback, self._config['pause'])
        # Update state.
        self._play_state = PLAY_STATE.PLAYING
        # Fire event to indicate content change.
        self.dispatch('on_content_change', self)

    def previous(self, reschedule=True):
        """Display previous file in index."""
        self.next(previous=True)

    def stop(self):
        """Stop playing slideshow."""
        # Skip if already stopped.
        if self._play_state == PLAY_STATE.STOPPED: return
        # Unschedule callback function.
        if self._next_event is not None:
            self._next_event.cancel()
            self._next_event = None
        # Remove current widget from layout.
        if self._current_widget is not None:
            self.remove_widget(self._current_widget)
        # Reset selective index iterator with sorting/filter criteria from the
        # slideshow configuration.
        self._iterator = None
        self._current_widget = None
        # Update state otherwise.
        self._play_state = PLAY_STATE.STOPPED
        # Fire event to indicate content change.
        self.dispatch('on_content_change', self)

    def _clock_callback(self, dt):
        """Clock callback function. Display the next file in the slideshow."""
        # Display next file in repository.
        self.next(reschedule=False)



"""Module providing slideshow content base widget."""

from kivy.clock import Clock
from kivy.graphics import Color, Rectangle
from kivy.uix.anchorlayout import AnchorLayout
from kivy.uix.label import Label
from kivy.uix.widget import Widget

from datetime import datetime
from math import ceil


class ContentBase(Widget):
    """Slideshow content base widget.

    Base class defining standard interface for slideshow content widgets.
    """

    def __init__(self, file, config):
        """Initialize slideshow content base instance."""
        super().__init__()
        self._file = file
        self._config = config

    @property
    def config(self):
        """Return content configuration.

        :return: configuration passed to the constructor
        :rtype: dict
        """
        return self._config

    @property
    def file(self):
        """Return linked repository file.

        Note that the value None may be returned in cases where no valid file
        is linked to the content widget.

        :return: linked repository file
        :rtype: repository.file
        """
        return self._file

    def play(self):
        """Start playing content."""
        pass

    def stop(self):
        """Stop playing content."""
        pass


class ErrorMessage(ContentBase):
    """Slideshow error message widget."""

    def __init__(self, message, config):
        """Initialize the slideshow message instance."""
        super().__init__(None, config)
        self._message = message
        self._bgcolor = config['bg_color']

        # Determine label color based on background color.
        lcolor = Color(*self._bgcolor)
        if lcolor.v > 0.5:
            lcolor.hsv = [ 0, 0, 0 ]
        else:
            lcolor.hsv = [ 0, 0, 1 ]

        # Create label for error message.
        self._label = Label(markup=True, halign="center", valign="center", color=lcolor.rgba, font_blended=True, text=message)
        self.add_widget(self._label)
        # Call update_canvas method after the widget's size has been set.
        self.bind(size=self.update_canvas)

    def update_canvas(self, *args):
        """Update canvas when the size of the widget changes.

        Fills canvas with the background color and adjusts the label position,
        size, font size and padding.
        """
        # Clear canvas and fill with background color.
        self.canvas.before.clear()
        with self.canvas.before:
            Color(*self._bgcolor)
            Rectangle(pos=(0, 0), size=self.size)

        # Position and resize label.
        self._label.pos = (self.x, self.y)
        self._label.size = (self.width, self.height)
        self._label.text_size = self._label.size
        # Set font size and padding of label.
        font_size = round(self.config.get('label_font_size', 0.05) * self.height)
        self._label.font_size = font_size
        padding = round(self.config.get('label_padding', 0.05) * self.width)
        self._label.padding = (padding, padding)

    @property
    def message(self):
        """Return message text.

        :return: message text
        :rtype: str
        """
        return self._message


class LabeledContent(ContentBase):
    """Slideshow labeled content base widget.

    Base class for slideshow image and video widgets. Provides basic
    functionality for labeling.
    """

    def __init__(self, file, config):
        """Initialize the labeled content instance."""
        super().__init__(file, config)

        self._events = list()

        # Create and add white (foreground) and black (shadow) labels.
        self._wlabel = Label(markup=True, halign="right", valign="bottom", color=(1, 1, 1, 1), font_blended=True)
        self._blabel = Label(markup=True, halign="right", valign="bottom", color=(0, 0, 0, 1))
        self.add_widget(self._blabel)
        self.add_widget(self._wlabel)
        # Set the label text.
        mode = config.get('label_mode', "off")
        if mode is True or mode == "on" or mode == "auto":
            label = self.label
            self._wlabel.text = label
            self._blabel.text = label
        # Call _adjust_label method after the widget's size has been set.
        self.bind(size=self.adjust_label)

#    def __del__(self):
#        """Delete the labeled content instance."""
#        # Cancel any clock events.
#        while self._events:
#            event = self._events.pop()
#            event.cancel()

    def adjust_label(self, *args):
        """Adjust label when the widget becomes visible and its size is set."""
        # Set font size.
        font_size = round(self.config.get('label_font_size', 0.05) * min(self.height, self.width))
        self._wlabel.font_size = font_size
        self._blabel.font_size = font_size
        # Set padding.
        padding = round(self.config.get('label_padding', 0.05) * min(self.height, self.width))
        self._wlabel.padding = (padding, padding)
        self._blabel.padding = (padding, padding)
        # Resize labels.
        offset = ceil(0.03*font_size)
        self._wlabel.pos = (self.x, self.y + offset)
        self._wlabel.size = (self.width - offset, self.height)
        self._wlabel.text_size = self._wlabel.size
        self._blabel.pos = (self.x + offset, self.y)
        self._blabel.size = (self.width - offset, self.height - offset)
        self._blabel.text_size = self._blabel.size
        # Schedule events to turn labels off and on.
        mode = self.config.get('label_mode', "on")
        pause = self.config.get('pause')
        duration = self.config.get('label_duration', 24)
        if mode == "auto" and pause is not None and 2*duration < pause:
            self._events.append(Clock.schedule_once(self.label_off, duration))
            self._events.append(Clock.schedule_once(self.label_on, pause - duration))

    def label_off(self, dt=0):
        "Turn label off."
        self._wlabel.text = ""
        self._blabel.text = ""

    def label_on(self, dt=0):
        "Turn label on."
        self._wlabel.text = self.label
        self._blabel.text = self.label

    @property
    def config(self):
        """Return content configuration.

        :return: configuration passed to the constructor
        :rtype: dict
        """
        return self._config

    @property
    def file(self):
        """Return linked repository file.

        :return: linked repository file
        :rtype: repository.file
        """
        return self._file

    @property
    def label(self):
        """Return label text.

        The label text is built from properties and meta data of the linked
        file. Text creation can be controlled via the configuration.

        :return: label text
        :rtype: str
        """
        label = str()
        label_content = self.config.get('label_content', "short")
        # Add description and separate if available.
        description = self.file.description
        if description:
            label= f"[b]{description}[/b] · "
        # Return if only description requested.
        if label_content == "description": return
        # Add shortened geopgraphical location (if available)
        location = self.file.location
        if location is not None:
            location, _, country = str.rpartition(location, ",")
            location, _, region = str.rpartition(location, ",")
            location, _, city = str.rpartition(location, ",")
            # Dirty hack to replace lengthy
            if city == "":
                label = label + f"{country.strip()} · "
            else:
                label = label + f"{city.strip()}, {country.strip()} · "
        # Format and append creation date.
        if label_content == "short":
            date_str = self.file.creation_date.strftime("%Y-%m-%d")
        else:
            date_str = self.file.creation_date.strftime("%Y-%m-%d %H:%M")
        label = label + f"{date_str}"
        # Return if only short label requested.
        if label_content == "short": return label
        # Format and append tags if any.
        if self.file.tags:
            label = label + " ·[i]" + "".join(f" #{tag}" for tag in self.file.tags) + "[/i]"
        # Append file and repository uuid.
        label = label + f" · {self.file.uuid} [i]in[/i] {self.file.rep.uuid}"
        return label

    def on_parent(self, *largs):
        """Unschedule all clock events upon loss of parent.

        Callback function for parent property change events. Cancels all
        scheduled clock events if widget is removed from parent, i.e. parent
        is set to None.
        """
        # Cancel any clock events.
        if self.parent is None:
            while self._events:
                event = self._events.pop()
                event.cancel()



"""MQTT interface of picframe."""

import json
import logging
import os
import paho.mqtt.client as mqtt
import ssl
import time

from kivy.clock import Clock
from kivy.logger import Logger

from repository import check_param, check_valid_required

from .controller import DISPLAY_MODE, DISPLAY_STATE, PLAY_STATE, Controller

from ..common import APPLICATION_NAME, APPLICATION_DESCRIPTION, VERSION, PROJECT_NAME


# Pyframe and Home Assistant root topics
ROOT_TOPIC = "pyframe"
HA_ROOT_TOPIC = "homeassistant"

def entity_id(name):
    """Convert entity name to entity id."""
    return name.lower().replace(' ', '_')

def unique_id(device_id, entity_id):
    """Combine device and entity id to unique id."""
    return f"{device_id}_{entity_id}"


class MqttInterface:
    """MQTT remote control.

    Implements an MQTT based remote control for pyframe photo frames. The
    remote control is automatically registered in home assistant via the MQTT
    discovery service.
    """

    # Required and valid configuration parameters
    CONF_REQ_KEYS = {'host', 'user', 'password'}
    CONF_VALID_KEYS = {'port', 'tls', 'tls_insecure', 'device_id', 'device_name'} | CONF_REQ_KEYS

    def __init__(self, config, controller):
        """ Initialize MQTT interface instance.

        Establishes a connection to the MQTT broker based on the provided
        configuration. Note that only password based authentication is supported
        at the moemnt. The controller instance is used to control the pyframe
        photo frame.

        The following configuration parameters are supported:
            host: MQTT broker (required)
            port: connection port (default: 8883)
            tls: true if a secure connection shall be used (default: true)
            tls_insecure: true if insecure TLS connections, i.e. connections
                with non-trusted certificates, are permitted (default: false)
            user: login name (required)
            password: login password (required)
            device_id: pyframe device id (default: "pyframe")
            device_name: pyframe device name (default: device_id)

        :param config: MQTT interface configuration
        :param controller: Pyframe controller
        :type config: dict
        :type controller: pyframe.controller
        :raises: ConfigError
        """
        self._config = config
        self._controller = controller
        self._client = None
        self._event = None
        self._connection_lost = False
        self._reconnect_time = time.time()

        # Check the configuration for valid and required parameters.
        check_valid_required(config, self.CONF_VALID_KEYS, self.CONF_REQ_KEYS)
        # Check paramaters
        check_param('host', config, is_str=True)
        check_param('port', config, required=False, is_int=True, gr=0)
        check_param('tls', config, required=False, is_bool=True)
        check_param('tls_insecure', config, required=False, is_bool=True)
        check_param('user', config, is_str=True)
        check_param('password', config, is_str=True)
        check_param('device_id', config, required=False, is_str=True)
        check_param('device_name', config, required=False, is_str=True)

        host = config['host']
        port = config.get('port', 8883)
        tls = config.get('tls', True)
        tls_insecure = config.get('tls_insecure', False)
        user = config['user']
        password = config['password']

        self._device_id = config.get('device_id', "01memories")
        self._device_name = config.get('device_name', self._device_id)
        self._availability_topic = f"homeassistant/switch/{self._device_id}/available"

        try:
            # Configure mqtt client.
            client = mqtt.Client(client_id = self._device_id, clean_session=True)
            self._client = client
            client.username_pw_set(user, password)
            if tls and tls_insecure is False:
                client.tls_set(cert_reqs=ssl.CERT_REQUIRED)
            elif tls and tls_insecure is True:
                client.tls_set(cert_reqs=ssl.CERT_NONE)
                client.tls_insecure_set(True)
            # Register callback functions.
            client.on_connect = self.on_connect
            client.on_disconnect = self.on_disconnect
            client.on_message = self.on_message
            # Make client attempt to reconnect up to 60 s.
            client.reconnect_delay_set(min_delay=1, max_delay=60)
            # Publish initial availabiliy as "offline".
            client.will_set(self._availability_topic, "offline", qos=0, retain=True)
            # Establish connection.
            Logger.info(f"MQTT: Connecting to broker '{host}:{port}' as user '{user}'.")
            client.connect(host, port, 60)
            # Call network loop every x seconds.
            self._event = Clock.schedule_interval(self.loop, 0.2)
        except Exception as e:
            raise Exception(f"An exception occured during setup of the connection. {e}.")

    def __setup_select(self, client, name, options, icon=None, category=None):
        """Helper function to setup selections in Home Assistant.

        :param client: MQTT client instance
        :type client: paho.mqtt.client
        :param name: name of selection
        :type name: str
        :param options: selectable options
        :type options: list of str
        :param icon: name of icon
        :type icon: str
        :param category: optional category name
        :type category: str
        """
        eid = entity_id(name)
        uid = unique_id(self._device_id, eid)
        config_topic = f"{HA_ROOT_TOPIC}/select/{uid}/config"
        state_topic = f"{ROOT_TOPIC}/{self._device_id}/state"
        command_topic = f"{ROOT_TOPIC}/{self._device_id}/{eid}/set"

        payload = {
            "name": name,
            "object_id": uid,
            "unique_id": uid,
            "options": options,
            "availability_topic": self._availability_topic,
            "state_topic": state_topic,
            "command_topic": command_topic,
            "value_template": "{{ value_json." + eid + "}}",
            "device": {
                "name": self._device_name,
                "identifiers": [self._device_id],
                "model": APPLICATION_NAME,
                "sw_version": VERSION,
                "manufacturer": PROJECT_NAME
            }
        }
        if icon:
            payload["icon"] = icon
        if category:
            payload["entity_category"] = category
        payload = json.dumps(payload)

        client.publish(config_topic, payload, qos=0, retain=True)
        client.subscribe(command_topic, qos=0)

    def __setup_sensor(self, client, name, icon, has_attributes=False, category=None):
        """Helper function to setup sensors in Home Assistant.

        :param client: MQTT client instance
        :type client: paho.mqtt.client
        :param name: name of sensor
        :type name: str
        :param icon: name of icon
        :type icon: str
        :param has_attributes: optional flag indicating whether setup has
            attributes (default: false)
        :type has_attributes: bool
        :param category: optional category name
        :type category: str
        """
        eid = entity_id(name)
        uid = unique_id(self._device_id, eid)
        config_topic = f"{HA_ROOT_TOPIC}/sensor/{uid}/config"
        state_topic = f"{ROOT_TOPIC}/{self._device_id}/{eid}/state"
        attributes_topic = f"{ROOT_TOPIC}/{self._device_id}/{eid}/attributes"

        payload = {
            "name": name,
            "object_id": uid,
            "unique_id": uid,
            "icon": icon,
            "availability_topic": self._availability_topic,
            "state_topic": state_topic,
            "value_template": "{{ value_json." + eid + "}}",
            "device": {
                "name": self._device_name,
                "identifiers": [self._device_id],
                "model": APPLICATION_NAME,
                "sw_version": VERSION,
                "manufacturer": PROJECT_NAME
            }
        }
        if has_attributes:
            payload["json_attributes_topic"] = attributes_topic
        if category:
            payload["entity_category"] = category
        payload = json.dumps(payload)

        client.publish(config_topic, payload, qos=0, retain=True)

    def __setup_button(self, client, name, icon, category=None):
        """Helper function to setup buttons in Home Assistant.

        :param client: MQTT client instance
        :type client: paho.mqtt.client
        :param name: name of sensor
        :type name: str
        :param icon: name of icon
        :type icon: str
        :param category: optional category name
        :type category: str
        """
        eid = entity_id(name)
        uid = unique_id(self._device_id, eid)
        config_topic = f"{HA_ROOT_TOPIC}/button/{uid}/config"
        command_topic = f"{ROOT_TOPIC}/{self._device_id}/{eid}/set"

        payload = {
            "name": name,
            "object_id": uid,
            "unique_id": uid,
            "icon": icon,
            "availability_topic": self._availability_topic,
            "command_topic": command_topic,
            "payload_press": "ON",
            "device": {
                "name": self._device_name,
                "identifiers": [self._device_id],
                "model": APPLICATION_NAME,
                "sw_version": VERSION,
                "manufacturer": PROJECT_NAME
            }
        }
        if category:
            payload["entity_category"] = category
        payload = json.dumps(payload)

        client.subscribe(command_topic, qos=0)
        client.publish(config_topic, payload, qos=0, retain=True)

    def loop(self, dt):
        """Loop function.

        Processes MQTT messages via the Paho MQTT client loop function. Called
        regularly from the Kivy loop (scheduled via Clock). In the beginning,
        the method checks for disconnects and if necessary attempts to reconnect
        to the broker. Reconnection attempts are limited to once every 60s.
        """
        client = self._client
        # Attempt to reconnect if connection has been lost. The flag is set by
        # the on_disconnect callback function.
        if self._connection_lost:
            # Skip iteration if last attempt was less than 60s ago.
            if (time.time() - self._reconnect_time) < 60: return
            Logger.info(f"MQTT: Attempting to reconnect to broker.")
            try:
                # Update time of last reconnect attempt.
                self._reconnect_time = time.time()
                # Attempt to reconnect.
                rc = client.reconnect()
                if rc != mqtt.MQTT_ERR_SUCCESS:
                    Logger.error(f"MQTT: Reconnection to broker failed with error code {rc}.")
                    return
                self._connection_lost = False
                # Ensure new state is published automatically after content change.
                self._controller.bind(on_state_change=self.publish_state)
            except Exception as e:
                Logger.error(f"MQTT: Reconnection to broker failed. {e}.")
                return
        # Process messages.
        try:
            client.loop(timeout=0)
        except Exception as e:
            Logger.error(f"MQTT: An exception occurred while processing messages. {e}.")

    def on_connect(self, client, userdata, flags, rc):
        """Update availability and setup sensors/controls after successful
        connection to broker.

        Callback function for connection events. See Paho MQTT client
        documentation for explanation of userdata, flags and rc parameters.

        :param client: MQTT client instance
        :type client: paho.mqtt.client
        :param userdata: not used
        :type userdata: none
        :param flags: not used
        :type flags: none
        :param rc: connection result code
        :type rc: int
        """
        # Verify successful connection.
        if rc != mqtt.MQTT_ERR_SUCCESS:
            Logger.error(f"MQTT: Connection to broker failed with error code {rc}.")
            return
        Logger.info("MQTT: Connection to broker established.")

        # Update availability.
        Logger.debug("MQTT: Changing availability to 'online'.")
        client.publish(self._availability_topic, "online", qos=0, retain=True)

        Logger.debug("MQTT: Creating control elements.")
        # Create buttons.
        self.__setup_button(client, "Play", "mdi:play")
        self.__setup_button(client, "Pause", "mdi:pause")
        self.__setup_button(client, "Stop", "mdi:stop")
        self.__setup_button(client, "Next", "mdi:skip-next")
        self.__setup_button(client, "Previous", "mdi:skip-previous")
        self.__setup_button(client, "Touch", "mdi:gesture-tap")
        # Create selections.
        self.__setup_select(client, "Display mode", [ item.value for item in DISPLAY_MODE ], category="config")
        self.__setup_select(client, "Display state", [ item.value for item in DISPLAY_STATE ], category="config")
        self.__setup_select(client, "Play state", [ item.value for item in PLAY_STATE ], category="config")
        self.__setup_select(client, "Slideshow", self._controller.slideshows, category="config")
        # Create sensors.
        self.__setup_sensor(client, "File", "mdi:file-image", has_attributes=True)

        # Publish the current state
        self.publish_state()
        # Ensure new state is published automatically after content change.
        self._controller.bind(on_state_change=self.publish_state)

    def on_disconnect(self, client, userdata, rc):
        """Reconnect to broker in case of unexpected disconnect.

        Callback function for disconnection events. See Paho MQTT client
        documentation for explanation of userdata and rc parameters.

        :param client: MQTT client instance
        :type client: paho.mqtt.client
        :param userdata: not used
        :type userdata: none
        :param rc: connection result code
        :type rc: int
        """
        self._controller.unbind(on_state_change=self.publish_state)
        if rc != mqtt.MQTT_ERR_SUCCESS:
            Logger.error(f"MQTT: Connection to broker lost with error code {rc}.")
            self._connection_lost = True

    def on_message(self, client, userdata, message):
        """Process messages from subscribed topics.

        Callback function message events. See Paho MQTT client documentation for explanation of userdata and and message parameters.

        :param client: MQTT client instance
        :type client: paho.mqtt.client
        :param userdata: not used
        :type userdata: none
        :param message: MQTT message
        :type message: dict
        """
        payload = message.payload.decode("utf-8")
        device_id = self._device_id

        def __topic(name):
            return f"{ROOT_TOPIC}/{self._device_id}/{entity_id(name)}/set"

        # Process button messages.
        if message.topic == __topic("Play"):
            if payload == "ON":
                Logger.debug("MQTT: 'Play' button was pressed.")
                self._controller.play()
        elif message.topic == __topic("Pause"):
            if payload == "ON":
                Logger.debug("MQTT: 'Pause' button was pressed.")
                self._controller.pause()
        if message.topic == __topic("Stop"):
            if payload == "ON":
                Logger.debug("MQTT: 'Stop' button was pressed.")
                self._controller.stop()
        elif message.topic == __topic("Previous"):
            if payload == "ON":
                Logger.debug("MQTT: 'Previous' button was pressed.")
                self._controller.previous()
        if message.topic == __topic("Next"):
            if payload == "ON":
                Logger.debug("MQTT: 'Next' button was pressed.")
                self._controller.next()
        elif message.topic == __topic("Touch"):
            if payload == "ON":
                Logger.debug("MQTT: 'Touch' button was pressed.")
                self._controller.touch()

        # Process selections messages.
        elif message.topic == __topic("Display mode"):
                Logger.debug(f"MQTT: 'Display mode' was changed to '{payload}'.")
                self._controller.display_mode = payload
        elif message.topic == __topic("Display state"):
                Logger.debug(f"MQTT: 'Display state' was changed to '{payload}'.")
                self._controller.display_state = payload
        elif message.topic == __topic("Play state"):
                Logger.debug(f"MQTT: 'Play state' was changed to '{payload}'.")
                self._controller.play_state = payload
        elif message.topic == __topic("Slideshow"):
                Logger.debug(f"MQTT: 'Slideshow' was changed to '{payload}'.")
                self._controller.slideshow = payload

    def publish_state(self, *largs):
        client = self._client

        # Update file sensor and attributes if current file available.
        if self._controller.current_file is not None:
            Logger.debug("MQTT: Updating file sensor and attributes.")
            file = self._controller.current_file
            eid = entity_id("File")
            uid = unique_id(self._device_id, eid)
            topic_head = f"{ROOT_TOPIC}/{self._device_id}/{eid}"
            # Update state.
            state_topic = f"{topic_head}/state"
            payload = json.dumps({ eid: file.uuid })
            client.publish(state_topic, payload, qos=0, retain=False)
            # Update attributes.
            attributes_topic = f"{topic_head}/attributes"
            payload = {
                'Description': file.description,
                'Location': file.location,
                'Creation date': file.creation_date.strftime("%Y-%m-%d %H:%M:%S"),
                'Last modified': file.last_modified.strftime("%Y-%m-%d %H:%M:%S"),
                'Rating': file.rating,
                'Tags': "".join(f"#{tag} " for tag in file.tags),
                'Size': f"{file.width} x {file.height}",
                'Repository': file.rep.uuid
            }
            payload = json.dumps(payload)
            client.publish(attributes_topic, payload, qos=0, retain=False)

        # Update selections.
        file = self._controller.current_file
        Logger.debug("MQTT: Updating configuration selections.")
        state_topic = f"{ROOT_TOPIC}/{self._device_id}/state"
        payload = {
            entity_id("Display mode"): self._controller.display_mode,
            entity_id("Display state"): self._controller.display_state,
            entity_id("Play state"): self._controller.play_state,
            entity_id("Slideshow"): self._controller.slideshow
        }
        payload = json.dumps(payload)
        client.publish(state_topic, payload, qos=0, retain=False)

    def stop(self):
        """Stop MQTT interface."""
        if self._event is not None:
            Clock.unschedule(self._event)
        if self._client is not None:
            self._client.disconnect()



""""Module providing repository file class."""

import exifread
import ffmpeg
import fnmatch
import locale
import logging

from abc import ABC, abstractmethod
from datetime import datetime
from geopy.geocoders import Photon
from geopy.exc import GeopyError
from iptcinfo3 import IPTCInfo

from PIL import Image


# Global geolocator instance for reverse location lookups.
geolocator = Photon()

class RepositoryFile:
    """File within a repository.

    Abstract base class providing basic functionality common to all File
    subclasses.

    Properties:
        uuid (str): Universally unique identifier (UUID) of the file. Typically
            the full path of the file within the repository.
        rep (repository.Repository): Repository containing the file.
        name (str): Name of the file. Default is None.
        index (repository.Index): Optional file metadata index. Default is
            None.
        type (int): Type of the file. Default is None. If set the following
            values may be returend:
            TYPE_UNKNOWN: Unknown file type
            TYPE_IMAGE: Image file
            TYPE_VIDEO: Video file
        width (int): Width of the file content in pixels. Default is None.
        height (int): Height of the file content in pixels. Default is None.
        rotation (int): Clock-wise rotation of the content in degrees. Default
            is None. If set the following values may be returned: 0, 90, 180
            or 270.
        orientation (int): Orientation of the content considering rotation.
            Default is None. If set the following values may be returned:
            ORIENTATION_LANDSCAPE: Content wider than high (default)
            ORIENTATION_PORTRAIT: Content heigher than wide
        creation_date (datetime): Creation date of file content.
        last_modified (datetime): Date of last file modification.
        last_updated (datetime): Date of last metadata update.
        description (str): Description of the file content. Default is None.
        rating (int): Rating of the file content. Default is None.
        coordinates (list of float): Geographical coordinates of location [ latitude,
          longitude, altitude ]. Default is [ None, None, None ].
        location (str): Human-readable location (address). Default is None.
        source (str): Source of the file (e.g. full path or URL).
        tags (set of str): User defined tags on the file. Not to be confounded
            with EXIF tags in image files. Default is None.
    """

    # File type definitions
    TYPE_UNKNOWN = 0
    TYPE_IMAGE = 1
    TYPE_VIDEO = 2

    # Orientation definitions
    ORIENTATION_LANDSCAPE = 0
    ORIENTATION_PORTRAIT = 1

    def __init__(self, uuid, rep, index=None, index_lookup=True):
        """Initialize file instance.

        :param uuid: UUID of the file.
        :type uuid: str
        :param rep: Repository containing file
        :type rep: repository.Repository
        :param index: Optional file metadata index. Default is None.
        :param index_lookup: True if file metadata shall be looked up from index.
        :type index_lookup: bool
        :type index: repository.Index
        """
        # Basic initialization.
        self._in_index = False
        self._uuid = uuid
        self._rep = rep
        self._index = index
        self._type = RepositoryFile.TYPE_UNKNOWN
        self._name = str()
        self._width = 0
        self._height = 0
        self._rotation = 0
        self._orientation = RepositoryFile.ORIENTATION_LANDSCAPE
        self._creation_date = datetime.today()
        self._last_modified = datetime.today()
        self._last_updated = datetime.today()
        self._description = str()
        self._rating = None
        self._coordinates = [ None, None, None ]
        self._location = None
        self._tags = list()

        # Attempt to determine type from extension.
        self._type_from_extension()

        # Try to retrieve metadata from index if available.
        mdata = None
        if index is not None and index_lookup is True:
            mdata = index.lookup(self, rep)

        if mdata is not None:
            logging.debug(f"Assigning metadata of file '{self._uuid}' from index.")
            self._in_index = True
            self._type = mdata.type
            self._width = mdata.width
            self._height = mdata.height
            self._rotation = mdata.rotation
            self._orientation = mdata.orientation
            self._creation_date = mdata.creation_date
            self._last_modified = mdata.last_modified
            self._last_updated = mdata.last_updated
            self._description = mdata.description
            self._rating = mdata.rating
            self._coordinates = [ mdata.latitude, mdata.longitude, mdata.altitude ]
            self._tags = [tag.name for tag in mdata.tags]

    def __repr__(self):
        """Provide string representation of file instance.

        :return: String representation
        :rtype: str
        """
        return f"File=(uuid='{self.uuid}', rep=0x{id(self.rep):x}, name={self.name}, type={self.type}, width={self.width}, height={self.height}, rotation={self.rotation}, orientation={self.orientation}, creation_date='{self.creation_date}', description='{self.description}', rating={self.rating}, tags={self.tags})"

    # Extensions of supported files
    EXT_IMAGE = ("*.jpg", "*.jpeg", "*.png")
    EXT_VIDEO = ("*.mp4", "*.mv4", "*.mov")

    def _extract_image_metadata(self, path):
        """Extract image metadata from file content.

        Uses the exifread library to extract EXIF metadata and IPTCInfo3
        library to extract IPTC metadata from the image file. Metadata is
        stored in the corresponding object properties.

        Note the star rating tag is not yet supported by exifread and therefore
        always remains at the default.
        """
        # Save current datetime in property last_updated.
        self._last_updated = datetime.today()

        # Use PIL to determine image size.
        with Image.open(path) as image:
            self._width, self._height = image.size

        # Open image file for reading (binary mode) and extract EXIF information.
        with open(path, 'rb') as file:
            tags = exifread.process_file(file)

        # Obtain rotation from metadata if available
        if 'Image Orientation' in tags:
            orientation = tags['Image Orientation'].values[0]
            if orientation == 8:
                self._rotation = 90
            elif orientation == 3:
                self._rotation = 180
            elif orientation == 6:
                self._rotation = 270

        # Derive orientation from dimensions and rotation.
        if (self._width < self._height and (self.rotation == 0 or self.rotation == 180)) or (self._width > self._height and (self.rotation == 90 or self.rotation == 270)):
            self._orientation = RepositoryFile.ORIENTATION_PORTRAIT
        else:
            self._orientation = RepositoryFile.ORIENTATION_LANDSCAPE

        # Etract image description if available.
        if 'Image ImageDescription' in tags:
            self._description = tags['Image ImageDescription'].values

        # Extract image rating if available.
        if 'Image Rating' in tags:
            self._rating = tags['Image Rating'].values[0]

        # Extract creation date if available.
        try:
            if 'EXIF DateTimeOriginal' in tags:
                creation_date = tags['EXIF DateTimeOriginal'].values
                self._creation_date = datetime.strptime(creation_date, "%Y:%m:%d %H:%M:%S")
            elif 'Image DateTime' in tags:
                creation_date = tags['Image DateTime'].values
                self._creation_date = datetime.strptime(creation_date, "%Y:%m:%d %H:%M:%S")
        except ValueError:
            logging.error(f"Invalid creation time format '{creation_date}'.")

        # Extract latitude and longitude from GPS data.
        if tags.keys() >= {'GPS GPSLatitude', 'GPS GPSLatitudeRef', 'GPS GPSLongitude', 'GPS GPSLongitudeRef'}:
            try:
                # Latitude
                values = tags['GPS GPSLatitude'].values
                latitude = float(values[0] + values[1]/60 + values[2]/3600)
                if tags['GPS GPSLatitudeRef'].values[0] == "S": latitude = -latitude
                self._coordinates[0] = latitude
                # Longitude
                values = tags['GPS GPSLongitude'].values
                longitude = float(values[0] + values[1]/60 + values[2]/3600)
                if tags['GPS GPSLongitudeRef'].values[0] == "W": longitude = -longitude
                self._coordinates[1] = longitude
            except Exception:
                logging.error(f"Invalid format of GPS coodinates.")

        # Extract altitude from GPS data.
        if 'GPS GPSAltitude' in tags:
            altitude = float(tags['GPS GPSAltitude'].values[0])
            self._coordinates[2] = altitude

        # Extract image tags(keywords) from IPTC tag if available
        info = IPTCInfo(path, force=True, inp_charset="utf8")
        self._tags = info["keywords"]

    def _extract_video_metadata(self, path):
        """Extract video metadata from file content.

        Uses the ffmpeg ffprobe command to extract video metadata from the
        video file and stores metadata in corresponding object properties.
        """
        # Save current datetime in property last_updated.
        self._last_updated = datetime.today()

        # Obtain meta data of all streams (video and audio).
        streams = ffmpeg.probe(path)['streams']
        logging.debug(streams)
        # Find first video stream and return corresponding meta data.
        for data in streams:
            if data.get('codec_type') == 'video': break

        # Try to obtain image dimensions from metadata.
        if data.get('width') is not None:
            self._width = int(data.get('width'))
        if data.get('height') is not None:
            self._height = int(data.get('height'))

        # Stop here if meta data does not contain any tags.
        if data.get('tags') is None: return

        # Try to obtain rotation from metadata.
        if data.get('side_data_list') is not None:
            rotation = data['side_data_list'][0].get('rotation')
            if rotation is not None:
                if rotation < 0: rotation = 360 + rotation
                self._rotation = rotation

        logging.debug(f"self._rotation: {self._rotation}")

        # Derive orientation from dimensions and rotation.
        if (self._width < self._height and (self.rotation == 0 or self.rotation == 180)) or (self._width > self._height and (self.rotation == 90 or self.rotation == 270)):
            self._orientation = RepositoryFile.ORIENTATION_PORTRAIT
        else:
            self._orientation = RepositoryFile.ORIENTATION_LANDSCAPE

        # Try to extract creation time from metadata.
        if data.get('tags') is not None:
            creation_date = data['tags'].get('creation_time')
            if creation_date is not None:
                try:
                    self._creation_date = datetime.strptime(creation_date, "%Y-%m-%dT%H:%M:%S.%fZ")
                except ValueError:
                    logging.error(f"Invalid creation time format {creation_date}.")

    def _type_from_extension(self):
        """Determine file type based on file extension."""
        fname = self._uuid.upper()
        for pattern in RepositoryFile.EXT_IMAGE:
            if fnmatch.fnmatch(fname, pattern.upper()):
                self._type = RepositoryFile.TYPE_IMAGE
                return
        for pattern in RepositoryFile.EXT_VIDEO:
            if fnmatch.fnmatch(fname, pattern.upper()):
                self._type = RepositoryFile.TYPE_VIDEO
                return

    @abstractmethod
    def extract_metadata(self):
        """Extract metadata from file content.

        Used to enforce the metadata extraction from file content in cases
        where not done automatically during initiation.
        """
        pass

    @property
    def uuid(self):
        """Return UUID of the file.

        :return: Universally unique identifier (UUID) of the file. Typically
            the full path of the file within the repository.
        :rtype: str
        """
        return self._uuid

    @property
    def rep(self):
        """Return repository containing the file.

        :return: Repository containing the file.
        :rtype: repository.Repository
        """
        return self._rep

    @property
    def name(self):
        """Return name of the file.

        :return: Name of the file.
        :rtype: str
        """
        return self._name

    @property
    def source(self):
        """Return the source of the file (e.g. full path or URL).

        :return: Source of the file.
        :rtype: str
        """
        return None

    @property
    def type(self):
        """Return type of the file.

        :return: Type of the file. The following values may be returend:
            TYPE_UNKNOWN: Unknown file type
            TYPE_IMAGE: Image file
            TYPE_VIDEO: Video file
        :rtype: int
        """
        return self._type

    @property
    def width(self):
        """Return width of the file content.

        :return: Width of the file content in pixels.
        :rtype: int
        """
        return self._width

    @property
    def height(self):
        """Return height of the file content.

        :return: Height of file content in pixels.
        :rtype: int
        """
        return self._height

    @property
    def rotation(self):
        """Return rotation of the file content.

        :return: Clock-wise rotation of the content in degrees. Typical
            values are 0, 90, 180 and 270.
        :rtype: int
        """
        return self._rotation

    @property
    def orientation(self):
        """Return orientation of the file content.

        :return: Orientation of the content considering rotation. The
            following values may be returned:
            ORIENTATION_LANDSCAPE: Content wider than high (default)
            ORIENTATION_PORTRAIT: Content heigher than wide.
        :rtype: int
        """
        return self._orientation

    @property
    def creation_date(self):
        """Return creation date of the file content.

        :return: Creation date of the file content. Note that this is not
            necessarily the creation date of the file.
        :rtype: DateTime
        """
        return self._creation_date

    @property
    def last_modified(self):
        """Return date of last file modification.

        :return: Date of last file modification.
        :rtype: DateTime
        """
        return self._last_modified

    @property
    def last_updated(self):
        """Return date of last metadata update.

        :return: Date of last metadata update.
        :rtype: DateTime
        """
        return self._last_updated

    @property
    def description(self):
        """Return description of the file content.

        :return: Description of the file content.
        :rtype: str
        """
        return self._description

    @property
    def rating(self):
        """Return rating of the file content.

        :return: Rating of the file content. Typically a star rating with values
            from 1 to 5.
        :rtype: int
        """
        return self._rating

    @property
    def coordinates(self):
        """Return coordinates where the file content was created.

        :return: [ latitude, longitude, height ]. Note that elements may be None if information is not available.
        :rtype: list of floats
        """
        return self._coordinates

    @property
    def location(self):
        """Return location (address) where the file content was created.

        May return None if the geographical location is not available.

        The geographical location is not stored as part of the file meta data.
        Since all free geocoding services implement a rate limit, reverse
        location lookups would slow down the indexing process. The geographical
        location is thus only looked up once requested.

        :return: city/country
        :rtype: str
        """
        # Return cached value if avaialable.
        if self._location is not None: return self._location
        # Return None if no coordinates available.
        if self._coordinates[0] is None or self._coordinates[1] is None:
            return None

        # Try to obtain location from geocoding service otherwise.
        try:
            # Retrieve default language from locale.
            lc = locale.getdefaultlocale()
            if len(lc) >= 1 and len(lc[0]) >= 2:
                lang = lc[0][0:2]
            else:
                lang = "en"
            self._location = geolocator.reverse(self._coordinates[0:2], exactly_one=True, language=lang).address
        except GeopyError as e:
            logging.error(f"An error occurred while looking up the location. {e}")
        finally:
            return self._location

    @property
    def tags(self):
        """Return tags on the file.

        :return: Tags on the file.
        :rtype: set of str
        """
        return self._tags



"""Module for WebDAV repository files."""

import logging
import os
import os.path
import repository
import tempfile

from datetime import datetime
from repository import IoError, UuidError


class RepositoryFile(repository.RepositoryFile):
    """File within a webdav repository.

    See repository.File for documentation of properties.
    """

    def __init__(self, uuid, rep, index=None, index_lookup=True, extract_metadata=True):
        """Initialize the repository file.

        :param rep: WebDAV repository
        :type rep: repository.webdav.Repository
        :param uuid: UUID of repository file
        :type uuid: str
        :param index: Optional file metadata index. Default is None.
        :type index: repository.Index
        :param index_lookup: Flag indicating whether file metadata shall be
            looked up from index. Default is True.
        :type index_lookup: bool
        :param extract_metadata: Flag indicating whether file metadata shall be
            extracted from file if not available from index. Default is True.
        :type extract_metadata: bool
        :raises: repository.UuidError, repository.IoError
        """
        # Call constructor of parent class.
        super().__init__(uuid, rep, index, index_lookup)

        # Basic initialization.
        self._cache_file = None
        self._path = None

        # Set file name from uuid
        self._name = os.path.basename(uuid)

        # Attempt to determine last modification and file creation date.
        if not self._in_index:
            # Attempt to retrieve file attributes.
            try:
                info = self._rep.client.info(self.uuid)
            except Exception as e:
                raise IoError(f"An exception occurred while retrieving WebDAV file information. {e}", e)
            # Attempt to determine last modified date.
            try:
                modified = info.get('modified')
                last_modified = datetime.strptime(modified, "%a, %d %b %Y %H:%M:%S %Z")
                self._last_modified = last_modified
            except (ValueError, TypeError):
                logging.warn(f"Failed to convert last modified date string '{modified}' to datetime.")
                last_modified = self.last_modified
            # Attempt to determine file creation date.
            try:
                created = info.get('created')
                creation_date = datetime.strptime(created, "%a, %d %b %Y %H:%M:%S %Z")
                self._creation_date = creation_date
            except (ValueError, TypeError):
                logging.warn(f"Failed to convert created date string '{created}' to datetime.")

        # Attempt to extract metadata from file content.
        if not self._in_index and extract_metadata:
            self.extract_metadata()

    def __del__(self):
        """Delete the file."""
        # Close and delete cache file if exists to prevent clean up errors.
        if self._cache_file is not None:
            try:
                self._cache_file.close()
            except FileNotFoundError:
                # Ignore any file not found errors, which can occur if the
                # cache directory and its content are deleted before the
                # temporary file.
                pass

    def _download(self):
        """Download file from WebDAV repository to local cache file.

        :raises: repository.IoError
        """
        if self._cache_file is None:
            # Create temporary file for local caching inside cache directory
            # of the WebDav repository.
            self._cache_file = tempfile.NamedTemporaryFile(dir=self._rep.cache_dir)
            self._path = self._cache_file.name
            logging.debug(f"Local cache file '{self._path}' created.")

            # Download file from WebDav repository.
            logging.info(f"Downloading file '{self.uuid}' from webdav repository to local cache file.")
            try:
                self._rep.client.download_from(self._cache_file, self._uuid)
            except Exception as e:
                raise repository.IoError(f"An exception occurred while downloading file '{self._uuid}' from WebDAV repository. {e}", e)

    def extract_metadata(self):
        """Extract metadata from file content."""
        # Download file if file type is supported.
        if self.type in (repository.RepositoryFile.TYPE_IMAGE, repository.RepositoryFile.TYPE_VIDEO):
            self._download()
        # Attempt to extract metadata from file content.
        logging.debug(f"Extracting metadata of file '{self.uuid}' from file content.")
        if self._type == repository.RepositoryFile.TYPE_IMAGE:
            self._extract_image_metadata(self._path)
        elif self._type == repository.RepositoryFile.TYPE_VIDEO:
            self._download()
            self._extract_video_metadata(self._path)

    @property
    def source(self):
        """Return full path of the local cache file.

        :return: full path
        :rtype: str
        """
        # Make sure to download file before returning path.
        self._download()
        # Return full path to local cache file.
        return self._path



"""Pyframe main module.

Author: Bernd Kalbfuss
License: GNU General Public License v3 (GPLv3)
"""

import argparse


# Parse arguments
parser = argparse.ArgumentParser(description='01 memories digital photo frame application.')
parser.add_argument('command', type=str, choices = ['show', 'index'])
parser.add_argument('items', type=str, nargs='*')
parser.add_argument('--rebuild', action='store_const', const=True, default=False)
args = parser.parse_args()


# Start slideshow application.
if args.command == 'show':
    from .show import run_app
    run_app()
# Start repository indexer.
elif args.command == 'index':
    from .index import run_indexer
    run_indexer(args.items, args.rebuild)
else:
    parser.print_help()



"""Collection of classes to display content in slideshows.

Rendering of the following content is currently supported:

* Images via :class:`pyframe.content.SlideshowImage`
* Videos via :class:`pyframe.content.SlideshowVideo`
* Error messages via :class:`pyframe.content.ErrorMessage

Author: Bernd Kalbfuß
License: GNU General Public License v3 (GPLv3)
"""

from . base import ContentBase, ErrorMessage, LabeledContent
# Use PIL for image loading and manipulation
from . pil_image import SlideshowImage
#from . image import SlideshowImage
from . video import SlideshowVideo



"""Module providing repository class"""

import logging

from abc import ABC, abstractmethod

from .common import ConfigError, UuidError


class Repository(ABC):
    """Repository of files.

    Abstract base class providing basic functionality common to all Repository
    sub-classes.
    """

    # Maximum length of uuid
    MAX_LEN_UUID = 36
    # Dictionary of all repositories by uuid
    _repositories = dict()

    # Required and valid configuration parameters. Need to be re-defined by
    # implementing sub-class.
    CONF_REQ_KEYS = {}
    CONF_VALID_KEYS = {}

    def __init__(self, uuid, config, index=None):
        """Initialize the repository.

        :param uuid: UUID of the repository.
        :type uuid: str
        :param config: Repository configuration from the configuration file.
        :type config: dict
        :param index: Optional file metadata index. Default is None.
        :type index: repository.Index
        :raises: UuidError
        """
        # Basic intialization.
        self._uuid = uuid
        self._index = index

        # Check the configuration for errors.
        self._check_config(config)

        # Test uuid for validity.
        if len(uuid) >= Repository.MAX_LEN_UUID:
            raise UuidError(f"UUID for repository too long. Maximum of {Repository.MAX_LEN_UUID} characters allowed.", uuid)
        # Warn if uuid already in use, but do not throw execption.
        if uuid in Repository._repositories:
            logging.warn(f"UUID for repository '{uuid}' is already in use.")

        # Add self to dictionary of repositories.
        Repository._repositories[uuid] = self

    def __del__(self):
        """Delete the repository."""
        if self.uuid in Repository._repositories:
            del Repository._repositories[self.uuid]

    @abstractmethod
    def iterator(self, index_lookup=True, extract_meta_data=True):
        """Provide iterator which allows to traverse through all files in the repository.

        :param index_lookup: True if file metadata shall be looked up from index.
        :type index_lookup: bool
        :return: File iterator.
        :return type: repository.FileIterator
        """
        pass

    def __iter__(self):
        """Provide iterator which allows to traverse through all files in the repository.

        :return: file iterator
        :return type: repository.FileIterator
        """
        return self.iterator()

    @abstractmethod
    def _check_config(self, config):
        """Check the configuration for the repository from the configuration
        file. This method is abstract and needs to be implemented by child
        classes.

        :param config: repository configuration
        :type config: dict
        :raises: ConfigError
        """
        pass

    @staticmethod
    def by_uuid(uuid):
        """Return an existing repository instance by its UUID. Raises a
        UuidError if no repository with the specified UUID exists.

        :param uuid: UUID of the repository.
        :type uuid: str
        :return: Repository with matching UUID.
        :rtype: repository.Repository
        :raises: UuidError
        """
        if uuid in Repository._repositories:
            return Repository._repositories[uuid]
        else:
            raise UuidError(f"There is no repository with UUID '{uuid}'", uuid)

    @abstractmethod
    def file_by_uuid(self, uuid, index_lookup=True, extract_metadata=True):
        """Return a file within the repository by its UUID. Raises a UuidError
        if the repository does not contain a file with the specified UUID.

        :param uuid: UUID of the file.
        :type uuid: str
        :param index_lookup: True if metadata shall be looked up from index.
        :type index_lookup: bool
        :param extract_metadata: True if metadata shall be extracted from the
          file if not available from the index.
        :type extract_metadata: bool
        :return: File with matching UUID.
        :rtype: repository.RepositoryFile
        :raises: UuidError
        """
        pass

    @property
    def index(self):
        """Return metadata index of the repository.

        :return: Metadata index of the repository. May return None if no index
            was specified.
        :rtype: repository.Index
        """
        return self._index

    @index.setter
    def index(self, index):
        """Set metadata index of the repository.

        :param index: Metadata index of the repository.
        :type index: repository.Index
        """
        self._index = index

    @staticmethod
    def repositories():
        """Return set of UUIDs of existing repositories.

        :return: UUIDs of existing repositories.
        :rtype: set
        """
        return set(Repository._repositories.keys())

    @property
    def uuid(self):
        """Return UUID of the repository.

        :return: UUID of repository.
        :rtype: str
        """
        return self._uuid


class FileIterator:
    """Iterator which can be used to traverse through files in a repository."""

    @abstractmethod
    def __next__(self):
        """Provide the next file.

        :returns: Next file in the repository.
        :rtype: repository.RepositoryFile
        :raises: StopIteration
        """
        pass

    def __iter__(self):
        """Provide self as iterator for traversing through all files in the
        repository.

        Enables the Repository.iterator method to create iterators with
        additional arguments.

        :return: File iterator.
        :return type: repository.FileIterator
        """
        return self



"""Collection of classes to access and index file repositories.

Provides interface definitions for classes :class:`repository.Repository` and :class:`repository.File`. Actual implementations are provided by sub-packages.

The following implementations, i.e. sub-packages, are currently available:
    - local: Files are stored on a local file system.
    - webdav: Files are stored on a WebDAV accessible share.

The :class:`repository.Index` class provides functionality to index file meta
data for the purpose of caching, filtering and sorting.

Author: Bernd Kalbfuss
License: GNU General Public License v3 (GPLv3)
"""

from .common import ConfigError, UuidError, IoError, check_valid_required, check_param
from .file import RepositoryFile
from .repository import Repository, FileIterator
from .index import SORT_DIR, SORT_ORDER, Index, MetaData



"""Module for rclone repositories."""

import logging
import repository
import tempfile

from rclone_python import rclone
from repository import ConfigError, IoError, check_param, check_valid_required

from .file import RepositoryFile


class Repository(repository.Repository):
    """Repository with rclone file base.

    The rclone repository class currently does not provide any functionality
    to configure rclone remotes. Remotes must have been configured before in
    the rclone configuration file, e.g. by using the *rclone config* command.

    The root directory needs to include the rclone remote (e.g.
    "owncloud:<myroot>").
    """

    # Required and valid configuration parameters
    CONF_REQ_KEYS = {'root', 'cache'}
    CONF_VALID_KEYS = set() | CONF_REQ_KEYS

    def __init__(self, uuid, config, index=None):
        """Initialize the repository.

        :param uuid: UUID of repository
        :type name: str
        :param config: dictionary with repository configuration
        :type config: dict
        :param index: optional file metadata index. Default is None.
        :type index: repository.Index
        :raises: repository.UuidError, repository.ConfigError
        """
        # Call constructor of parent class.
        repository.Repository.__init__(self, uuid, config, index)
        # Basic initialization.
        self._root = config.get('root', "/")
        # Create temporary directory for file caching.
        self._cache_dir = tempfile.TemporaryDirectory(dir=config['cache'], prefix=f"{uuid}-")

    def _check_config(self, config):
        """Check the repository configuration.

        :param config:
        :type config: dict
        :raises: repository.ConfigError
        """
        # Make sure all valid and required parameters have been specified.
        check_valid_required(config, self.CONF_VALID_KEYS, self.CONF_REQ_KEYS)
        # Check parameter values.
        check_param('root', config, is_str=True)
        check_param('cache', config, is_str=True)

    @property
    def cache_dir(self):
        """Return path of cache directory for the temporary storage of files.

        :return: path of cache directory
        :rtype: str
        """
        return self._cache_dir.name

    def iterator(self, index_lookup=True, extract_metadata=True):
        """Provide iterator to traverse through files in the repository.

        :param index_lookup: Flag indicating whether file metadata shall be
            looked up from index. Default is True.
        :type index_lookup: bool
        :param extract_metadata: Flag indicating whether file metadata shall be
            extracted from file if not available from index. Default is True.
        :type extract_metadata: bool
        :return: file iterator
        :return type: repository.FileIterator
        """
        return FileIterator(self, index_lookup, extract_metadata)

    def file_by_uuid(self, uuid, index_lookup=True, extract_metadata=True):
        """Return file within the repository by its UUID.

        Raises an IoError if the file with UUID does not exist/cannot be
        accessed.

        :param uuid: UUID of repository file
        :type uuid: str
        :param index_lookup: Flag indicating whether file metadata shall be
            looked up from index. Default is True.
        :type index_lookup: bool
        :param extract_metadata: Flag indicating whether file metadata shall be
            extracted from file if not available from index. Default is True.
        :type extract_metadata: bool
        :return: file with matching UUID
        :rtype: repository.RepositoryFile
        :raises: repository.IoError
        """
        return RepositoryFile(uuid, self, self._index, index_lookup, extract_metadata)

    @property
    def root(self):
        """Return root directory of the repository.

        :return: root directory
        :rtype: str
        """
        return self._root


class FileIterator(repository.FileIterator):
    """Iterator to traverse through files in an rclone repository."""

    def __init__(self, rep, index_lookup=True, extract_metadata=True):
        """Initialize file iterator.

        :param rep: rclone repository
        :type rep: repository.rclone.Repository
        :param index_lookup: Flag indicating whether file metadata shall be
            looked up from index. Default is True.
        :type index_lookup: bool
        :param extract_metadata: Flag indicating whether file metadata shall be
            extracted from file if not available from index. Default is True.
        :type extract_metadata: bool
        :raises: repository.IoError
        """
        # Basic initialization.
        self._rep = rep
        self._index_lookup = index_lookup
        self._extract_metadata = extract_metadata

        # Create iterator for list of files in root directory.
        try:
            # The default max_depth of rclone is 1. A value of 0 menas no limit,
            # but is not passed on by the wrapper. As workaround we specify a
            # reasonably high number.
            self._file_list = rclone.ls(f"{rep.root}", max_depth=1000)
        except Exception as e:
            raise IoError(f"An exception occurred while listing the root directory. {e}", e)
        self._iterator = iter(self._file_list)

    def __next__(self):
        """Provide next file in iteration.

        :returns: next file
        :rtype: repository.rclone.RepositoryFile
        :raises: StopIteration
        """
        # Retrieve the next directory entry.
        entry = self._iterator.__next__()
        # Skip all directories.
        while entry['IsDir']:
            logging.debug(f"Skipping directory '{entry['Path']}'.")
            # Save all sub-directories for later.
            entry = self._iterator.__next__()

        # Derive uuid from path.
        uuid = entry['Path']
        # Return the next file.
        logging.debug(f"Creating rclone repository file '{uuid}'.")
        return self._rep.file_by_uuid(uuid, self._index_lookup, self._extract_metadata)



"""Module providing common definitions."""

import copy
import logging
import os
import os.path
import sys
import yaml

from importlib import import_module
from logging.handlers import TimedRotatingFileHandler
from kivy.logger import Logger
from repository import check_param, ConfigError, Index, Repository, UuidError

#from .mylogging import Handler


APPLICATION_NAME = "Pyframe"
APPLICATION_DESCRIPTION = "Digital photo frame"
VERSION = "0.x"
PROJECT_NAME = "Pyframe project"

# Default configuration
DEFAULT_CONFIG = {
    'bg_color': [1, 1, 1],
    'cache': os.path.expanduser(f"~/.cache/01memories/cache"),
    'direction': "ascending",
    'display_mode': "static",
    'display_state': "on",
    'display_timeout': 300,
    'enable_animation': True,
    'enable_exception_handler': False,
    'enable_logging': True,
    'enable_scheduler': True,
    'enable_mqtt': True,
    'index': os.path.expanduser(f"~/.cache/01memories/index.sqlite"),
    'index_update_interval': 0,
    'label_mode': "off",
    'label_content': "full",
    'label_duration': 60,
    'label_font_size': 0.08,
    'label_padding': 0.03,
    'log_level': "warning",
    'log_dir': os.path.expanduser(f"~/.cache/01memories/log"),
    'max_zoom': 1.3,
    'pause': 300,
    'resize': "fill",
    'rotation': 0,
    'smart_limit': 10,
    'smart_time': 24,
    'sound': True,
    'order': "name",
    'window_position': "auto",
    'window_size': "full"
}


# Mapping of name to numeric log level.
LOG_LEVELS = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'critical': logging.CRITICAL
}


class Formatter(logging.Formatter):
    """Pyframe log formatter.

    Used to imitate the Kivy log format in rotating log files.
    """

    def __init__(self, *args):
        """Initiaize Formatter instance."""
        super().__init__(*args)

    def format(self, record):
        """Split and format record."""
        try:
            msg = record.msg.split(':', 1)
            if len(msg) == 2:
                record.msg = '[%-13s]%s' % (msg[0], msg[1])
        except:
            pass
        return super().format(record)


def _configure_logging(config, filename):
    """Configure logging.

    Adjusts log levels based on the application configuration and adds a
    custom log handler for logging to rotating log files.

    :param config: Application configuration
    :type config: dict
    :param filename: Log filename
    :type filename: str
    :raises: Raises an exception if the directory cannot be created or is
      not writable.
    """

    # Check parameters.
    check_param('enable_logging', config, is_bool=True)
    check_param('log_level', config, options=set(LOG_LEVELS.keys()))
    check_param('log_dir', config, is_str=True)

    # Set log levels of default python and Kivy Logger.
    numeric_level = LOG_LEVELS[config['log_level']]
    logging.getLogger().setLevel(numeric_level)
    Logger.setLevel(numeric_level)

    # Reduce logging by IPTCInfo and exifread to errors or specified log
    # level, whatever is higher.
    logging.getLogger("iptcinfo").setLevel(max(logging.ERROR, numeric_level))
    logging.getLogger("exifread").setLevel(max(logging.ERROR, numeric_level))
    # Reduce logging by SQLAlchemy to warnings or specified log level,
    # whatever is higher.
    logging.getLogger("sqlalchemy").setLevel(max(logging.WARN, numeric_level))

    # Create log directory if it does not exist yet.
    log_dir = config['log_dir']
    if not os.path.exists(log_dir):
        try:
            os.makedirs(log_dir, exist_ok=True)
        except Exception as e:
            raise Exception(f"An exception occurred while creating the log directory '{log_dir}': {e}")

    # Make sure the directory is writable.
    if not os.access(log_dir, os.W_OK):
        raise Exception(f"The log directory '{log_dir}' is not writeable.")

    # Write all log messages to a rotating log file.
    if config['enable_logging'] == "on" or config['enable_logging'] == True:
        try:
            fullpath = os.path.join(log_dir, filename)
            logHandler = TimedRotatingFileHandler(os.path.join(fullpath), when="h", interval=24, backupCount=5, encoding='utf-8', errors='ignore')
            # Apply Kivy style formatter
            formatter = Formatter("%(asctime)s [%(levelname)-8s] %(message)s", "%Y-%m-%d %H:%M:%S")
            logHandler.setFormatter(formatter)
            # Add rotating log file handler to default logger.
            logging.info(f"Enabling logging to file '{fullpath}'.")
            logging.getLogger().addHandler(logHandler)
        except Exception as e:
            raise Exception(f"An error occurred while installing the log handler: {e}")


def _create_repositories(config, index):
    """Create file repositories from configuration.

    :param config: Application configuration
    :type config: dict
    :param index: Repository index
    :type index: repository.Index
    :raises: ConfigError, Exception
    """

    # Dictionary used to map type names to repository classes.
    supported_types = {
        'local': ("repository.local", "Repository"),
        'webdav': ("repository.webdav", "Repository"),
        'rclone': ("repository.rclone", "Repository")}

    # Exit application if no repositories have been defined.
    if 'repositories' not in config or type(config['repositories']) is not dict:
        raise ConfigError("Exiting application as no repositories have been defined.")

    # Extract global repository index configuration.
    global_index_config = {key: config[key] for key in ('index_update_interval', 'index_update_at') if key in config}

    # Create repositories based on the configuration.
    for uuid, rep_config in config['repositories'].items():

        # Skip disabled repositories.
        enabled_flag = rep_config.get('enabled', True)
        if enabled_flag is False or enabled_flag == "off":
            logging.info(f"Configuration: Skipping repository '{uuid}' as it has been disabled.")
            continue

        # Check parameters.
        check_param('type', rep_config, options=set(supported_types.keys()))

        # Substitute patterns in root paths of local directories. The
        # following patterns are substituted:
        #   {sys.prefix} => sys.prefix (e.g. "/usr/local")
        if rep_config.get('type') == "local":
            rep_config['root'] = os.path.expanduser(rep_config.get('root'))
            rep_config['root'] = rep_config.get('root').replace("{sys.prefix}", sys.prefix, 1)

        # Retrieve repository class from type.
        ref = supported_types[rep_config.get('type')]
        module = import_module(ref[0])
        rep_class = getattr(module, ref[1])

        # Combine global and local repository configuration. Local
        # configuration settings supersede global settings.
        rep_config2 = copy.deepcopy(config)
        rep_config2.update(rep_config)
        rep_config = { key: rep_config2[key] for key in rep_class.CONF_VALID_KEYS if key in rep_config2 }

        try:
            # Create repository instance.
            logging.info(f"Configuration: Creating {rep_config2.get('type')} repository '{uuid}'.")
            rep = rep_class(uuid, rep_config, index=index)
        # Catch any invalid configuration and UUID errors.
        except (ConfigError, UuidError) as e:
            raise ConfigError(f"Error in the configuration of repository '{uuid}'. {e}", rep_config)

    # Raise exception if no valid repositories have been defined.
    if len(Repository._repositories.items()) == 0:
        raise ConfigError("Exiting application as no valid repositories have been defined.", config['repositories'])

    # Create cache directory if it does not exist yet.
    cache_dir = config['cache']
    try:
        os.makedirs(cache_dir, exist_ok=True)
    except Exception as e:
        raise Exception(f"An exception occurred while creating the cache file directory '{cache_dir}': {e}")


def _load_config():
    """Load application configuration.

    Loads the application configuration from the default configuration file
    and applies default values where missing.

    :returns: Application configuration
    :rtype: dict
    """
    conf_paths = [
        "./config.yaml",
        os.path.expanduser("~/.config/01memories/config.yaml"),
        "/etc/01memories/config.yaml",
        os.path.expanduser("~/.local/share/01memories/config/config.yaml"),
        f"{sys.prefix}/share/01memories/config/config.yaml",
        "../config/config-dev.yaml"
    ]

    # Determine path of configuration file.
    for path in conf_paths:
        if os.path.isfile(path): break

    # Load configuration from yaml file.
    logging.info(f"Configuration: Loading configuration from file '{path}'.")
    with open(path, 'r', encoding='utf8') as config_file:
        config2 = yaml.safe_load(config_file)

    # Copy and update default configuration.
    config = copy.deepcopy(DEFAULT_CONFIG)
    config.update(config2)

    logging.debug(f"Configuration: Configuration = {config}")
    return config


def _load_index(config):
    """Load or create repository index.

    :param config: Application configuration
    :type config: dict
    :returns: Repository index
    :rtype: repository.Index
    :raises: Exception
    """
    # Create index directory if it does not exist yet.
    index_path = config['index']
    index_dir = os.path.dirname(index_path)
    try:
        os.makedirs(index_dir, exist_ok=True)
    except Exception as e:
        raise Exception(f"An exception occurred while creating the index file directory '{index_dir}': {e}")

    # Load/create and return index.
    return Index(index_path)



"""Module providing slideshow image class."""

from kivy.graphics import PushMatrix, PopMatrix, Rotate, Color, Rectangle
from kivy.logger import Logger
from kivy.uix.image import Image
from kivy.uix.label import Label
from kivy.uix.widget import Widget

from .base import LabeledContent


class SlideshowImage(LabeledContent):
    """Image slideshow widget.

    Loads the image from the specified File and starts playing it as soon as the
    widget becomes visible. The image is scaled to fit the entire widget,
    respecting the aspect ratio.
    """

    def __init__(self, file, config):
        """Initialize slideshow image instance.

        :param file: Repository file instance for the image to be displayed.
        :type file: repository.File
        :param config: Dictionary with the following entries:
            rotation: Angle in degrees (int) by which the image is rotated clockwise.
            bgolor: Canvas background color (list(3)) for areas, which are not covered by the image.
            resize: Mode (str) for resizing of images. Must equal "fit" or "fill".
        :type config: dict
        """
        super().__init__(file, config)
        # Calculate total rotation from image rotation and rotation configured for the slideshow.
        # To avoid negative values and values beyond 360°, we additionally apply a modulo operation.
        self._rotation = (file.rotation - config['rotation']) % 360
        self._bgcolor = config['bg_color']
        self._resize = config['resize']
        # Create and add image widget
        self._image = Image(source=file.source, mipmap=True)
        self.add_widget(self._image, len(self.children))
        # Call update_canvas method when the size of the widget changes.
        self.bind(size=self.update_canvas)

    def update_canvas(self, *args):
        """Update canvas when the size of the widget changes."""
        # Clear before and after groups of image canvas.
        self._image.canvas.before.clear()
        self._image.canvas.after.clear()

        # Fill canvas with background color.
        with self._image.canvas.before:
            Color(*self._bgcolor)
            Rectangle(pos=(0, 0), size=self.size)

        # Determine aspect ratios of image slideshow widget (this widget)
        # and image.
        widget_ratio = self.width/self.height
        image_ratio = self._image.image_ratio
        # Correct image aspect ratio for image rotation. i.e. aspect ratio
        # corresponds to the ratio after rotation.
        if abs(self._rotation) == 90 or abs(self._rotation == 270):
            image_ratio = 1/image_ratio

        # Tranform image to fill the image slideshow widget. Only images with
        # the same orientation will be resized to fill the widget. Images with a
        # different orientation will be resized to fit the widget.
        if self._resize == "fill":

            # Determine maximum dimension for widget with landscape orientation.
            if widget_ratio > 1:
                # Determine required maximum dimension for the rotation
                # transformation based on aspect ratios.
                if widget_ratio > image_ratio and image_ratio > 1:
                    max_dim = self.width
                elif widget_ratio <= image_ratio and image_ratio >= 1:
                    max_dim = round(self.height*image_ratio)
                elif widget_ratio >= image_ratio and image_ratio <= 1:
                    max_dim = self.height
                else:  # widget_ratio < image_ratio and image_ratio < 1
                    max_dim = round(self.width/image_ratio)
            # Determine maximum dimension for widget with portrait orientation.
            else:  # widget_ratio <= 1:
                if widget_ratio > image_ratio and image_ratio > 1:
                    max_dim = round(self.height*image_ratio)
                    # max_dim = self.width
                elif widget_ratio <= image_ratio and image_ratio >= 1:
                    max_dim = self.width
                    # max_dim = round(self.height*image_ratio)
                elif widget_ratio >= image_ratio and image_ratio <= 1:
                    # max_dim = self.height
                    max_dim = round(self.width/image_ratio)
                else:  # widget_ratio < image_ratio and image_ratio < 1
                    # max_dim = round(self.width/image_ratio)
                    max_dim = self.height

            # Set size of image widget to square with maximum dimension
            self._image.size = (max_dim, max_dim)
            # Adjust position of image widget within slideshow image widget
            # to center rotated image.
            self._image.x = round(self.x + (self.width - max_dim)/2)
            self._image.y = round(self.y + (self.height - max_dim)/2)

            # Apply rotation if not zero
            if self._rotation != 0:
                with self._image.canvas.before:
                    PushMatrix()
                    Rotate(angle=self._rotation, origin=self._image.center, axis=(0, 0, 1))
                with self._image.canvas.after:
                    PopMatrix()

        # Default is to fit the image to the canvas
        else:  # self._resize == "fit"

            # Resize and rotate the image if required.
            if self._rotation != 0:
                # Determine required maximum dimension for the rotation
                # transformation based on aspect ratios.
                if widget_ratio > image_ratio and image_ratio > 1:
                    max_dim = round(self.height*image_ratio)
                elif widget_ratio <= image_ratio and image_ratio >= 1:
                    max_dim = self.width
                elif widget_ratio >= image_ratio and image_ratio <= 1:
                    max_dim = self.height
                else:  # widget_ratio < image_ratio and image_ratio < 1
                    max_dim = round(self.width/image_ratio)

                # Set size of image widget to square with maximum dimension
                self._image.size = (max_dim, max_dim)
                # Adjust position of image widget within slideshow image widget
                # to center rotated image.
                self._image.x = round(self.x + (self.width - max_dim)/2)
                self._image.y = round(self.y + (self.height - max_dim)/2)

                # Apply rotation.
                with self._image.canvas.before:
                    PushMatrix()
                    Rotate(angle=self._rotation, origin=self._image.center, axis=(0, 0, 1))
                with self._image.canvas.after:
                    PopMatrix()

            # Set size of image widget to size of image slideshow widget (this
            # widget) otherwise and let image widget do the scaling.
            else:  # self._rotation == 0:
                self._image.size = self.size

            # Log debug information
#            Logger.debug(f"Image uuid: {self._file.uuid}")
#            Logger.debug(f"Image type: {self._file.type}")
#            Logger.debug(f"Image source: {self._file.source}")
#            Logger.debug(f"Image orientation: {self._file.orientation}")
#            Logger.debug(f"Image rotation: {self._file.rotation}")
#            Logger.debug(f"Total rotation: {self._rotation}")
#            Logger.debug(f"Widget width: {self.width}")
#            Logger.debug(f"Widget height: {self.height}")
#            Logger.debug(f"Widget aspect ratio: {widget_ratio}")
#            Logger.debug(f"max_dim: {max_dim}")
#            Logger.debug(f"Image width: {self._image.width}")
#            Logger.debug(f"Image height: {self._image.height}")
#            Logger.debug(f"Image aspect ratio: {image_ratio}")
#            Logger.debug(f"Image x: {self._image.x}")
#            Logger.debug(f"Image y: {self._image.y}")
#            Logger.debug(f"Image center: {self._image.center}")



"""Module providing Pyframe controller interface."""

from abc import ABC, abstractmethod
from enum import Enum


class DISPLAY_MODE(str, Enum):
    MOTION = "motion"
    STATIC = "static"

class DISPLAY_STATE(str, Enum):
    OFF = "off"
    ON = "on"

class PLAY_STATE(str, Enum):
    PAUSED = "paused"
    PLAYING = "playing"
    STOPPED = "stopped"


class Controller(ABC):
    """Pyframe controller.

    Provides interface for the control of photo frames. The class is abstract
    and needs to be implemented by the inheriting class.
    """

    @property
    @abstractmethod
    def current_file(self):
        """Return the current repository file.

        :return: current file
        :rtype: repository.File
        """
        pass

    @property
    @abstractmethod
    def display_mode(self):
        """Return display mode.

        See enumeration DISPLAY_MODE for possible values.

        :return: display mode
        :rtype: str
        """
        pass

    @display_mode.setter
    @abstractmethod
    def display_mode(self, mode):
        """Set display mode.

        See enumeration DISPLAY_MODE for possible values.

        :param mode: display mode
        :type mode: str
        """
        pass

    @abstractmethod
    def display_on(self):
        """Turn display on."""
        pass

    @abstractmethod
    def display_off(self):
        """Turn display off."""
        pass

    @property
    @abstractmethod
    def display_state(self):
        """Return display state.

        See enumeration DISPLAY_STATE for possible values.

        :return: display state
        :rtype: str
        """
        pass

    @display_state.setter
    @abstractmethod
    def display_state(self, state):
        """Set display state.

        See enumeration DISPLAY_STATE for possible values.

        :param state: display state
        :type mode: str
        """
        pass

    @property
    @abstractmethod
    def display_timeout(self):
        """Return display timeout.

        :return: display timeout in seconds
        :rtype: int
        """
        pass

    @display_timeout.setter
    @abstractmethod
    def display_timeout(self, timeout):
        """Set display timeout.

        :param timeout: display timeout in seconds
        :type timeout: int
        """
        pass

    @abstractmethod
    def next(self):
        """Change to next file in slideshow."""
        pass

    @abstractmethod
    def pause(self):
        """Pause playing the current slideshow."""
        pass

    @abstractmethod
    def play(self):
        """Start playing the current slideshow."""
        pass

    @property
    @abstractmethod
    def play_state(self):
        """Return play state.

        See enumeration PLAY_STATE for possible values.

        :return: play state
        :rtype: str
        """
        pass

    @play_state.setter
    @abstractmethod
    def play_state(self, state):
        """Set play state.

        See enumeration PLAY_STATE for possible values.

        :param mode: play state
        :type mode: str
        """
        pass

    @abstractmethod
    def previous(self):
        """Change to previous file in slideshow."""
        pass

    @property
    @abstractmethod
    def slideshow(self):
        """Return name of the current slideshow.

        :return: name of slideshow
        :rtype: str
        """
        pass

    @slideshow.setter
    @abstractmethod
    def slideshow(self, name):
        """Set current slideshow by its name.

        :param name: name of slideshow
        :type name: str
        """
        pass

    @property
    @abstractmethod
    def slideshows(self):
        """Return names of all slideshows.

        :return: list of slideshow names
        :rtype: list of str
        """
        pass

    @abstractmethod
    def stop(self, force=False):
        """Stop playing the current slideshow."""
        pass

    @abstractmethod
    def touch(self):
        """Touch to prevent screen timeout."""
        pass



"""Collection of classes to generate slideshows.

Author: Bernd Kalbfuss
License: GNU General Public License v3 (GPLv3)
"""

from .app import run_app
#from .common import APPLICATION_NAME, APPLICATION_DESCRIPTION, VERSION, PROJECT_NAME
#from .mylogging import Handler
#from .indexer import Indexer
#from .controller import Controller, DISPLAY_MODE, DISPLAY_STATE, PLAY_STATE
#from .scheduler import Scheduler
#from .slideshow import Slideshow
#from .mqtt import MqttInterface
#from .app import App



"""Module providing Pyframe slideshow application."""

import copy
import kivy.app
import logging
import os.path
import repository.local
import repository.webdav
import subprocess
import signal
import sys
import time
import traceback
import yaml

#from importlib import import_module
from repository import ConfigError, Index, Repository, UuidError, check_param, check_valid_required

from kivy.base import ExceptionManager, stopTouchApp
from kivy.core.window import Window
from kivy.clock import Clock
from kivy.logger import Logger, LOG_LEVELS

from .slideshow import Slideshow
from .scheduler import Scheduler
from .controller import Controller, DISPLAY_MODE, DISPLAY_STATE, PLAY_STATE
from .mqtt import MqttInterface

from ..common import _create_repositories, _configure_logging, _load_config, _load_index


app = None


def signal_handler(sig, frame):
    """Close application after SIGINT and SIGTERM signals."""
    global app
    Logger.warning(f"App: Signal '{signal.strsignal(sig)}' received. Preparing for safe exit.")
    app.close()
    stopTouchApp()


def run_app():
    """Start slideshow."""
    global app
    # Catch interrupt and term signals and exit gracefully.
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    # Run slideshow.
    app = App()
    app.run()
    # Clean up and exit.
    app.close()
    stopTouchApp()


class ExceptionHandler(kivy.base.ExceptionHandler):
    """Pyframe exception handler.

    Logs all exceptions, but continues with the execution. The main purpose of
    the handler is to prevent the application from exiting unexpectedly.
    """

    def __init__(self, app):
        """Initialize exception handler instance."""
        super().__init__()
        self._app = app

    def handle_exception(self, exception):
        """Log all exceptions and continue with execution.

        For unknown reasons, the exception is not always provided as an argument
        albeit the kivy documentation says so. The method thus retrieves the
        last raised exception via a system call.
        """
        # Make sure we do not produce any unhandled exceptions within the
        # exception handler.
        try:
            # Log information on exception.
            Logger.error("App: An exception was raised:")
            # Formatting the excpetion has failed for unknown reasons in the
            # past. Putting into separate try block to make sure subsequent
            # code is executed.
            try:
                # Retrieve information on last exception via system call since
                # exception object passed to handler sometimes seems to be
                # missing/incomplete.
                _, e, _ = sys.exc_info()
                Logger.error(f"App: {''.join(traceback.format_exception(e)).rstrip()}")
            except:
                pass
            Logger.error("App: Ignoring and continuing with execution.")
            # Wait for a moment to slow down infinite loops.
            time.sleep(10)
            # Restart slideshow if playing.
            if self._app.play_state == PLAY_STATE.PLAYING:
                self._app.stop()
                self._app.play()
            # Continue with execution.
        except: pass
        return ExceptionManager.PASS


class App(kivy.app.App, Controller):
    """Pyframe slideshow application."""

    # Required and valid configuration parameters
    CONF_REQ_KEYS = {'display_mode', 'display_state', 'display_timeout', 'enable_animation', 'enable_exception_handler', 'enable_mqtt', 'enable_logging', 'enable_scheduler', 'index', 'log_level', 'log_dir', 'max_zoom', 'repositories', 'slideshows', 'window_position', 'window_size'} | Slideshow.CONF_REQ_KEYS
    CONF_VALID_KEYS = {'cache', 'index_update_at', 'index_update_interval', 'mqtt', 'schedule', 'sound' } | CONF_REQ_KEYS | Slideshow.CONF_VALID_KEYS


    def _create_slideshows(self):
        """Create slideshows from configuration.

        Slideshow configurations are extracted from the 'slideshows' section
        in the configuration file. One slideshow is created per has entry.
        Slideshow instances are collected in the hash _slideshows, with the key
        being identical to the slideshow name in the configuration file.

        :raises: ConfigError
        """
        config = self._config
        index = self._index
        # Exit application if no slideshow has been defined.
        if 'slideshows' not in config or type(config['slideshows']) is not dict:
            raise ConfigError("Configuration: Exiting application as no slideshows have been defined.", config)

        # Create empty dictionary to collect slideshows
        self._slideshows = dict()
        # Extract global slideshow configuration
        global_config = {key: config[key] for key in Slideshow.CONF_VALID_KEYS if key in config}

        # Create slideshows from configuration.
        for name, config in config['slideshows'].items():
            # Combine global and local configuration. Local configuration
            # settings supersede global settings.
            combined_config = copy.deepcopy(global_config)
            combined_config.update(config)

            # Create new slideshow, bind 'on_content_change' event and add to
            # slideshow do dictionary.
            try:
                # Verify that only existing/enabled repositories have been defined.
                check_param('repositories', combined_config, required=False, recurse=True, options=Repository.repositories())
                # Create slideshow and add to the list of slideshows.
                slideshow = Slideshow(name, index, combined_config)
                self._slideshows[name] = slideshow
                # Make sure we receive all content change events.
                slideshow.bind(on_content_change=self.on_content_change)

            except ConfigError as e:
                raise ConfigError(f"Error in the configuration of slideshow '{name}'. {e}", config)

        # Exit application if no valid slideshows have been defined.
        if len(self._slideshows.items()) == 0:
            raise ConfigError("Exiting application as no valid slideshows have been defined.")


    def _init_display(self):
        """Initialize display and window.

        :raises: ConfigError
        """
        config = self._config

        # Check parameters.
        check_param('display_mode', config, options={ item.value for item in DISPLAY_MODE })
        check_param('display_state', config, is_bool=True)
        check_param('display_timeout', config, is_int=True, gr=0)

        # Convert from boolean to "on" (True) and "off" (False) if necessary.
        # This is a pecularity of the YAML 1.1 standard, which interprets "on"
        # and "off" as boolean values.
        map = {True: "on", False: "off", "on": "on", "off": "off"}
        display_state = map[config['display_state']]

        # Initialize display timeout, state and mode.
        self._display_timeout = config['display_timeout']
        self._timeout_event = None
        self._display_mode = ""
        self._display_state = ""
        self.display_state = display_state
        self.display_mode = config['display_mode']

        # Set window position.
        value = config['window_position']
        if type(value) is list and len(value) == 2 and value[0] >= 0 and value[1] >= 0:
            Window.left = value[0]
            Window.top = value[1]
        elif value == "auto":
            pass
        else:
            raise ConfigError(f"Invalid value '{value}' for parameter 'window_position' specified. Valid values are [left, top] and 'auto'.", config)

        # Set window size.
        value = config['window_size']
        if type(value) is list and len(value) == 2 and value[0] > 0 and value[1] > 0:
            Window.size = (value)
        elif value == "full":
            Window.fullscreen = 'auto'
        else:
            raise ConfigError(f"Invalid value '{value}' for parameter 'window_size' specified. Valid values are [width, height] and 'full'.", config)
        # Disable display of mouse cursor
        Window.show_cursor = False


    def build(self):
        """Build Kivy application.

        Loads the application configuration from the default configuration file.
        Creates configured repositories and builds an index across the latter.
        """
        self._index = None
        self._scheduler = None
        self._mqtt_interface = None
        self._play_state = PLAY_STATE.STOPPED

        # Register 'state_change' event, which is fired upon content and
        # controller state changes.
        self.register_event_type('on_state_change')

        # Load configuration.
        self._config = _load_config()
        # Check the configuration for valid and required parameters.
        check_valid_required(self._config, self.CONF_VALID_KEYS, self.CONF_REQ_KEYS)

        # Configure logging.
        _configure_logging(self._config, "pyframe.log")
        # Load/create index.
        self._index = _load_index(self._config)
        # Create repositories from configuration.
        _create_repositories(self._config, self._index)
        # Create slideshows.
        self._create_slideshows()

        # Make first slideshow the main root widget.
        self.root = next(iter(self._slideshows.values()))

        # Create mqtt interface if configured and activated.
        value = self._config.get('enable_mqtt')
        if 'mqtt' in self._config and (value == "on" or value is True):
            try:
                self._mqtt_interface = MqttInterface(self._config['mqtt'], self)
            except ConfigError as e:
                raise ConfigError(f"Error in the MQTT interface configuration. {e}", e.config)
            except Exception as e:
                raise Exception(f"MQTT: {e}")

        # Initialize display
        self._init_display()

        # Bind keyboard listener
        Window.bind(on_keyboard=self.on_keyboard)

        # Create scheduler if configured and activated.
        value = self._config.get('enable_scheduler')
        if 'schedule' in self._config and (value == "on" or value is True):
            try:
                self._scheduler = Scheduler(self._config['schedule'], self)
            except ConfigError as e:
                raise ConfigError(f"Configuration: {e}", e.config)
            except Exception as e:
                raise Exception(f"Scheduler: {e}")
        # Start playing first defined slideshow otherwise.
        else:
            self.play()

        # Install default exception handler to prevent the application from
        # exiting unexpectedly. All exceptions are caught and logged, but the
        # appplication continues with the execution afterwards.
        value = self._config.get('enable_exception_handler')
        if value == "on" or value is True:
            ExceptionManager.add_handler(ExceptionHandler(self))

        return self.root


    def on_content_change(self, slideshow, *largs):
        """Handle slideshow content change events."""
        Logger.debug(f"App: Event 'on_content_change' from slideshow '{slideshow.name}' received. Forwarding as event 'on_state_change'.")
        # Forward as 'on_state_change' event.
        self.dispatch('on_state_change')
        # Consume event.
        return True

    def on_keyboard(self, window, key, *args):
        """Handle keyboard events.

        The following events are currently supported:
        - Right arrow: Show net file.
        - Left arrow: Show previous file (not yet implemented).
        - Escape: Exit application.
        """
        Logger.info(f"App: Key '{key}' pressed.")
        # Exit application if escape key pressed.
        if key == 27:
            # Let the default handler do the necessary work.
            return False
        # Touch controller to prevent screen timeout.
        self.touch()
        # Display next file if right arrow pressed.
        if key == 275:
            self.next()
        # Display previous file if left arrow pressed.
        elif key == 276:
            self.previous()
        # Consume event.
        return True

    def on_state_change(self, *largs):
        """Default handler for 'on_state_change' events."""
        pass

    def close(self):
        """Prepare application for safe exit.

        Shuts down the MQTT remote controller, scheduler and stops the current
        the current slideswhow."""
        # Stop current slideshow.
        self.stop()
        # Stop MQTT interface if running.
        if self._mqtt_interface is not None:
            Logger.info("App: Stopping MQTT interface.")
            self._mqtt_interface.stop()
        # Stop scheduler if running.
        if self._scheduler is not None:
            Logger.info("App: Stopping scheduler.")
            self._scheduler.stop()
        # Close metadata index if open.
        if self._index is not None:
            Logger.info("App: Closing metadata index.")
            self._index.close()

    def on_display_timeout(self, dt):
        """Handle display timeouts in motion mode."""
        # Pause playing slideshow and turn display off.
        Logger.info(f"Controller: Display has timed out.")
        self.pause()
        self.display_off()

    @property
    def current_file(self):
        """Return the current repository file.

        :return: current file
        :rtype: repository.File
        """
        return self.root.current_file

    @property
    def display_mode(self):
        """Return display mode.

        See enumeration DISPLAY_MODE for possible values.

        :return: display mode
        :rtype: str
        """
        return self._display_mode

    @display_mode.setter
    def display_mode(self, mode):
        """Set display mode.

        See enumeration DISPLAY_MODE for possible values.

        :param mode: display mode
        :type mode: str
        """
        # Return if already in the right mode.
        if mode == self._display_mode: return
        Logger.info(f"Controller: Changing display mode from '{self._display_mode}' to '{mode}'.")
        # Turn display on and start playing slideshow.
        # Cancel previously scheduled timeout event.
        if self._timeout_event is not None:
            self._timeout_event.cancel()
        #  Set display to motion mode and start playing slideshow.
        if mode == DISPLAY_MODE.STATIC: pass
        elif mode == DISPLAY_MODE.MOTION:
            # Update last touch time stamp.
            self._last_touch = time.time()
            # Schedule schedule timeout event if not stopped.
            if self.play_state != PLAY_STATE.STOPPED:
                self._timeout_event = Clock.schedule_once(self.on_display_timeout, self._display_timeout)
        # Raise exception upon invalid display mode.
        else:
            raise Exception(f"The selected display mode '{mode}' is invalid. Acceptable values are '{[ item.value for item in DISPLAY_MODE ]}'.")
        # Update display mode.
        self._display_mode = mode
        self.dispatch('on_state_change')

    def display_on(self):
        """Turn the display on."""
        # Return if already on.
        if self._display_state == DISPLAY_STATE.ON: return
        Logger.info("Controller: Turning display on.")
        # Turn display on on Linux with X server.
        subprocess.run("/usr/bin/xset dpms force on", shell=True,  stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL)
        # Raise window to top (just in case) and make full screen again.
#        Window.raise_window()
#        if self._config['window_size'] == "full":
#            Window.fullscreen = 'auto'
        # Update display state.
        self._display_state = DISPLAY_STATE.ON
        self.dispatch('on_state_change')

    def display_off(self):
        """Turn the display off."""
        # Return if already off.
        if self._display_state == DISPLAY_STATE.OFF: return
        Logger.info("Controller: Turning display off.")
        # Turn display off on Linux with X server.
        subprocess.run("/usr/bin/xset dpms force off", shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        # Update display state.
        self._display_state = DISPLAY_STATE.OFF
        self.dispatch('on_state_change')

    @property
    def display_state(self):
        """Return display state.

        See enumeration DISPLAY_STATE for possible values.

        :return: display state
        :rtype: str
        """
        return self._display_state

    @display_state.setter
    def display_state(self, state):
        """Set display state.

        See enumeration DISPLAY_STATE for possible values.

        :param state: display state
        :type mode: str
        """
        if self._display_state == state: return
        if state == DISPLAY_STATE.ON: self.display_on()
        elif state == DISPLAY_STATE.OFF: self.display_off()
        else:
            raise Exception(f"The selected display state '{state}' is invalid. Acceptable values are '{[ item.value for item in DISPLAY_STATE ]}'.")

    @property
    def display_timeout(self):
        """Return display timeout.

        :return: display timeout in seconds
        :rtype: int
        """
        return self._display_timeout

    @display_timeout.setter
    def display_timeout(self, timeout):
        """Set display timeout.

        :param timeout: display timeout in seconds
        :type timeout: int
        """
        Logger.info(f"Controller: Setting display timeout to {timeout} s.")
        self._display_timeout = timeout

    def pause(self):
        """Pause playing the current slideshow."""
        # Skip if already paused or stopped.
        if self._play_state == PLAY_STATE.PAUSED or self._play_state == PLAY_STATE.STOPPED: return
        Logger.info(f"Controller: Pausing slideshow '{self.slideshow}'.")
        self.root.pause()
        self._play_state = PLAY_STATE.PAUSED
        self.dispatch('on_state_change')

    def play(self):
        """Start/resume playing the current slideshow."""
        # Skip if already playing.
        if self._play_state == PLAY_STATE.PLAYING: return
        Logger.info(f"Controller: Playing/resuming slideshow '{self.slideshow}'.")
        # Start playing current slideshow.
        self.root.play()
        self._play_state = PLAY_STATE.PLAYING
        self.display_on()
        self.touch()
        self.dispatch('on_state_change')

    @property
    def play_state(self):
        """Return play state.

        See enumeration PLAY_STATE for possible values.

        :return: play state
        :rtype: str
        """
        return self._play_state

    @play_state.setter
    def play_state(self, state):
        """Set play state.

        See enumeration PLAY_STATE for possible values.

        :param mode: play state
        :type mode: str
        """
        if self._play_state == state: return
        if state == PLAY_STATE.PLAYING: self.play()
        elif state == PLAY_STATE.PAUSED: self.pause()
        elif state == PLAY_STATE.STOPPED: self.stop()
        else:
            raise Exception(f"The selected play state '{state}' is invalid. Acceptable values are '{[ item.value for item in PLAY_STATE ]}'.")

    def stop(self):
        """Stop playing the current slideshow."""
        # Skip if already stopped.
        if self._play_state == PLAY_STATE.STOPPED: return
        Logger.info(f"Controller: Stopping slideshow '{self.slideshow}'.")
        # Cancel previously scheduled timeout event.
        if self._timeout_event is not None:
            self._timeout_event.cancel()
        # Stop playing current slideshow.
        self.root.stop()
        self._play_state = PLAY_STATE.STOPPED
        self.display_off()
        self.dispatch('on_state_change')

    def previous(self):
        """Change to previous file in slideshow."""
        # Skip if not playing.
        if self._play_state == PLAY_STATE.STOPPED: return
        Logger.info(f"Controller: Changing to previous file in slideshow '{self.slideshow}'.")
        self.root.previous()

    def next(self):
        """Change to next file in slideshow."""
        # Skip if not playing.
        if self._play_state == PLAY_STATE.STOPPED: return
        Logger.info(f"Controller: Changing to next file in slideshow '{self.slideshow}'.")
        self.root.next()

    @property
    def slideshow(self):
        """Return name of the current slideshow.

        :return: slideshow name
        :rtype: str
        """
        return self.root.name

    @slideshow.setter
    def slideshow(self, name):
        """Set current slideshow by its name.

        :param name: slideshow name
        :type name: str
        """
        # Retrieve slideshow by its name. Stick to the current slideshow if
        # specified slideshow does not exist.
        new_root = self._slideshows.get(name, self.root)
        if new_root is not self.root:
            # Save play state and stop playing the current slideshow.
            cur_play_state = self.play_state
            self.stop()
            # Replace the root widget.
            Logger.info(f"Controller: Setting slideshow to '{name}'.")
            Window.add_widget(new_root)
            Window.remove_widget(self.root)
            self.root = new_root
            self.play_state = cur_play_state
            self.dispatch('on_state_change')

    @property
    def slideshows(self):
        """Return names of all slideshows.

        :return: list of slideshow names
        :rtype: list of str
        """
        return list(self._slideshows.keys())

    def touch(self):
        """Update last touch time stamp and prevent screen timeout in display motion mode."""
        # Update last touch time stamp.
        self._last_touch = time.time()
        # Return if display not in motion mode or slideshow stopped.
        if self.display_mode != DISPLAY_MODE.MOTION or self.play_state == PLAY_STATE.STOPPED: return
        # Cancel previously scheduled timeout event.
        if self._timeout_event is not None:
            self._timeout_event.cancel()
        # Restart playing the slideshow.
        self.play()
        # Schedule new timeout event.
        self._timeout_event = Clock.schedule_once(self.on_display_timeout, self._display_timeout)
        # Log next display timeout.
        next_timeout_asc = time.asctime(time.localtime(self._last_touch + self._display_timeout))
        Logger.debug(f"Controller: Controller has been touched. Next display timeout scheduled in {self._display_timeout} s at {next_timeout_asc}.")



"""Module providing Pyframe repository indexer."""

import logging
import os
import signal
import sys

from repository import ConfigError, Index, Repository

from ..common import _create_repositories, _configure_logging, _load_config, _load_index


indexer = None

def signal_handler(sig, frame):
    """Close application after SIGINT and SIGTERM signals."""
    logging.warning(f"Indexer: Signal '{signal.strsignal(sig)}' received. Preparing for safe exit.")
    indexer.close()
    sys.exit(1)


def run_indexer(uuids, rebuild):
    """Run indexer."""
    # Catch interrupt and term signals and exit gracefully.
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    # Disable loading of GL4ES library.
    #del os.environ['LD_LIBRARY_PATH']
    # Run file indexer.
    indexer = Indexer(uuids, rebuild)
    indexer.run()
    # Clean up and exit.
    indexer.close()


class Indexer:
    """Pyframe repository indexer."""


    def __init__(self, uuids, rebuild):
        """Initialize repository indexer instance."""
        # Load configuration.
        self._config = _load_config()
        # Configure logging.
        _configure_logging(self._config, "indexer.log")

        # Load/create index.
        self._index = _load_index(self._config)
        # Create repositories from configuration.
        _create_repositories(self._config, self._index)

        # Assume all repositories shall be indexed if none specified.
        if not uuids: uuids = Repository.repositories()
        # Identify repositories based on uuid.
        self._reps = [ Repository.by_uuid(uuid) for uuid in uuids ]

        # Safe argumenst for later use.
        self._uuids = uuids
        self._rebuild = rebuild


    def run(self):
        """Run indexer."""
        logging.info(f"Indexer: Indexing the following repositories: {[ rep.uuid for rep in self._reps ]}")
        if self._rebuild:
            logging.info(f"Indexer: Index is rebuilt from scratch.")

        # Build index for selected repositories.
        for rep in self._reps:
            self._index.build(rep, rebuild=self._rebuild)


    def close(self):
        """Prepare application for safe exit."""
        # Close metadata index if open.
        if self._index is not None:
            logging.info("Indexer: Closing metadata index.")
            self._index.close()



"""Module providing common definitions."""


class ConfigError(Exception):
    """Pyframe configuration error.

    Generic configuration error excption class, which can be used by any
    configurable pyframe component.
    """

    def __init__(self, msg, config=None):
        """Initialize configuration error instance.

        :param config: invalid configuration (default: None)
        :type config: dict
        """
        super().__init__(msg)
        self.config = config


class UuidError(Exception):
    """Invalid universal unique identifier (UUID) error."""

    def __init__(self, msg, uuid=None):
        super().__init__(msg)
        self.uuid = uuid


class IoError(Exception):
    """Input/Outout error."""

    def __init__(self, msg, e):
        super().__init__(msg)
        self.exception = e


def check_valid_required(config, valid_keys, required_keys):
    """Check for valid and required configuration parameters.

    Checks whether only valid and all required parameters (keys) have been
    specified. Raises a configuration error exception otherwise.

    :param config: configuration to be checked
    :type config: dict
    :param valid_keys: valid configuration keys
    :type valid keys: set
    :param required_keys: required configuration keys
    :type required_keys: set
    :raises: ConfigError
    """
    # Make sure only valid parameters have been specified.
    keys = set(config.keys())
    if not keys.issubset(valid_keys):
        raise ConfigError(f"The configuration contains additional parameters. Only the parameters {valid_keys} are accepted, but the additional parameter(s) {keys.difference(valid_keys)} has/have been specified.", config)

    # Make sure all required parameters have been specified.
    if not required_keys.issubset(keys):
        raise ConfigError(f"As a minimum, the parameters {required_keys} are required, but the parameter(s) {required_keys.difference(keys)} has/have not been specified.")

def check_param(name, value, recurse=False, required=True, is_int=False, is_bool=False, is_str=False, is_color=False, is_time=False, gr=None, ge=None, lo=None, le=None, options=None):
    """Check validity of configuration parameter.

    Checks the validity of a configuration parameter based on specified
    tests. Raises a configuration error exception if acceptance criteria are
    not met.

    The parameter value may either be provided directly or indirectly
    in the form of a dictionary. If a dictionary is provided as value, the
    parameter value is looked up via the parameter name.

    The function allows to recurse into lists and sets, for instance if the
    parameter value is a list of strings or numbers. In this case, specified
    tests are applied to each element of the set/list.

    :param name: parameter name
    :type name: str
    :param value: parameter value or dictionary
    :type value: type of parameter value or dict
    :param recurse: True if function shall recurse into sets and lists
    :type recurse: bool
    :param required: True if parameter must exist in dictionary
    :type required: bool
    :param is_int: True if value must be integer
    :type is_int: bool
    :param is_bool: True if value must be boolean
    :type is_bool: bool
    :param is_time: True if value must be a non-empty string
    :type is_time: bool
    :param is_color: True if value must be valid color definition
    :type is_color: bool
    :param is_time: True if value must be valid time definition
    :type is_time: bool
    :param gr: parameter value must be > gr
    :type gr: numeric
    :param ge: parameter value must be >= ge
    :type ge: numeric
    :param lo: parameter value must be < lo
    :type lo: numeric
    :param le: parameter value must be <= le
    :type le: numeric
    :param options: valid parameter values
    :type options: set or list
    :raises: ConfigError
    """
    config = value
    # Obtain parameter value from dictionary if dictionary specified.
    if isinstance(value, dict):
        if name in config:
            value = config.get(name)
            # Check for non-type and empty lists/sets.
            if value is None or (isinstance(value, (list,set)) and len(value) ==0):
                raise ConfigError(f"Parameter '{name}' does not have a value.", config)
        else:
            if required is True: raise ConfigError(f"No value for required parameter '{name}' specified.", config)
            else: return

    # Recurse into list items if requested.
    if isinstance(value, (list,set)):
        if recurse:
            for v in value:
                check_param(name, v, False, required, is_int, is_bool, is_str, is_color, is_time, gr, ge, lo, le, options)
        # Prevent all further tests.
        return

    # Compile generic error message prefix.
    prefix = f"Invalid value '{value}' for parameter '{name}' specified."

    # Ensure that value is boolean.
    if is_bool:
        if not (isinstance(value, bool) or value == "on" or value == "off"):
            raise ConfigError(f"{prefix} Value must be boolean (true or false).", config)
        # Prevent all further tests.
        return

    # Ensure that value is a valid, non-empty string.
    if is_str:
        # Check string format.
        if not isinstance(value, str) or len(value) == 0:
            raise ConfigError(f"{prefix} Value must be a non-empty string.", config)
        # Prevent all further tests.
        return

    # Ensure that value is valid color definition.
    if is_color:
        if not isinstance(value, list) or len(value) != 3:
            raise ConfigError(f"{prefix} Value must be valid color definition of type [r, g, b].", config)
        for rgb in value:
            if not isinstance(rgb, (int, float)) or rgb < 0 or rgb > 1:
                raise ConfigError(f"{prefix} Color values must be numeric within range [0,1].", config)
        # Prevent all further tests.
        return

    # Ensure that value is a valid time string.
    if is_time:
        # Check string format.
        if not isinstance(value, str) or not re.match("[0-2][0-9]:[0-5][0-9]", value):
            raise ConfigError(f"{prefix} Value must be valid time definition of type [HH:MM].", config)
        # Verify that hour and minute values are within range.
        time = [int(x) for x in value.split(":")]
        if not (time[0] >= 0 and time[0] <= 23 and time[1] >= 0 and time[1] <= 59):
            raise ConfigError(f"{prefix} Hour or minute value is out of bounds.", config)
        # Prevent all further tests.
        return

    # Ensure that value is integer.
    if is_int and not isinstance(value, int):
        raise ConfigError(f"{prefix} Value must be integer.", config)
    # Ensure that value is greater than a certain value.
    if gr is not None and not value > gr:
        raise ConfigError(f"{prefix} Value must be > {gr}.", config)
    # Ensure that value is greater than or equal to a certain value.
    if ge is not None  and not value >= ge:
        raise ConfigError(f"{prefix} Value must be >= {ge}.", config)
    # Ensure that value is lower than a certain value.
    if lo is not None  and not value < lo:
        raise ConfigError(f"{prefix} Value must be < {lo}.", config)
    # Ensure that value is lower than or equal to a certain value.
    if le is not None  and not value <= le:
        raise ConfigError(f"{prefix} Value must be <= {le}.", config)
    # Ensure that value is one of pre-defined options.
    if options is not None and value not in options:
        raise ConfigError(f"{prefix} Valid values are {options}.", config)



"""Module for local repository files."""

import logging
import os.path
import repository

from repository import UuidError
from datetime import datetime


class RepositoryFile(repository.RepositoryFile):
    """File within a local repository.

    For images, the following properties are extracted from EXIF/IPTC tags if
    available: width, height, rotation, creation_date, description, rating, GPS
    coordinates, and tags (keywords).

    See repository.File for documentation of properties.
    """

    def __init__(self, uuid, rep, index=None, index_lookup=True, extract_metadata=True):
        """Initialize the repository file.

        :param rep: local repository
        :type rep: repository.local.Repository
        :param uuid: UUID of repository file
        :type uuid: str
        :param index: Optional file metadata index. Default is None.
        :type index: repository.Index
        :param index_lookup: Flag indicating whether file metadata shall be
            looked up from index. Default is True.
        :type index_lookup: bool
        :param extract_metadata: Flag indicating whether file metadata shall be
            extracted from file if not available from index. Default is True.
        :type extract_metadata: bool
        :raises: repository.UuidError, repository.IoError
        """
        # Call constructor of parent class.
        super().__init__(uuid, rep, index, index_lookup)

        # Throw exception if file does not exist.
        self._path = os.path.join(rep.root, uuid)
        if not os.path.isfile(self._path):
            raise UuidError("There is no file with UUID '{uuid}'.", uuid)

        # Set file name from uuid.
        self._name = os.path.basename(uuid)

        # Determine last modification and file creation date.
        last_modified = datetime.fromtimestamp(os.path.getmtime(self._path))
        if not self._in_index or self.last_updated < last_modified:
            self._last_modified = last_modified
            self._creation_date = datetime.fromtimestamp(os.path.getctime(self._path))

        # Attempt to extract metadata from file content.
        if (not self._in_index or self.last_updated < last_modified) and extract_metadata:
            self.extract_metadata()

    def extract_metadata(self):
        """Extract metadata from file content."""
        logging.debug(f"Extracting metadata of file '{self.uuid}' from file content.")
        # If image try to extract metadata from EXIF tag.
        if self._type == repository.RepositoryFile.TYPE_IMAGE:
            self._extract_image_metadata(self._path)
        elif self._type == repository.RepositoryFile.TYPE_VIDEO:
            self._extract_video_metadata(self._path)

    @property
    def source(self):
        """Return full path of the file.

        :return: full path
        :rtype: str
        """
        return self._path



"""Module providing metadata index class.

The following test script is based on the tutorial by Philipp Wagner [1]. The
script depends on the following debian packages:
- python3-sqlalchemy

References;
----------
1. https://bytefish.de/blog/first_steps_with_sqlalchemy/
"""

import logging
import random
import time

from enum import Enum
from sqlalchemy import asc, create_engine, desc, event, func, update, delete, or_, Column, DateTime, Float, ForeignKey, Integer, String, Boolean
from sqlalchemy.engine import Engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import backref, relationship, sessionmaker, scoped_session

from .common import UuidError, check_param, check_valid_required
from .file import RepositoryFile
from .repository import Repository


# Install listener for connection events to automatically enable foreign key
# constraint checking by SQLite.
@event.listens_for(Engine, "connect")
def set_sqlite_pragma(conn, *largs):
    """Enable foreign key constraint checking and write-ahead logging (WAL)
    in SQLite."""
#    logging.debug("Enable foreign key database constraints.")
    cursor = conn.cursor()
    cursor.execute("PRAGMA journal_mode=WAL")
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()

"""
# Install listener to record time prior to execution of statements.
@event.listens_for(Engine, "before_cursor_execute")
def before_cursor_execute(conn, *largs):
    conn.info.setdefault("query_start_time", []).append(time.time())

# Install listener to calculate and log total execution time.
@event.listens_for(Engine, "after_cursor_execute")
def after_cursor_execute(conn, *largs):
    total = time.time() - conn.info["query_start_time"].pop(-1)
    logging.debug(f"SQL query executed within {round(total*1000, 2)} ms.")
"""

Base = declarative_base()


class MetaData(Base):
    """Database model for file metadata.

    Database representation of file metadata using SQLAlchemy object relational
    mapper. Properties resemble properties of class repository.RepositoryFile.

    Properties:
        id(Integer): Numerical unique identifier. Automatically generated.
        rep_uuid(String(36)): Universally unique identifier of the repository
            containing the file.
        file_uuid(String(255)): Universally unique identifier of the file.
        type(Integer): Type of the file. See repository.RepositoryFile for
            acceptable values.
        width(Integer): Width of the file content in pixels.
        height(Integer): Height of the file content in pixels.
        rotation(Iteger): Clock-wise rotation of the content in degrees. See
            repository.RepositoryFile for acceptable values.
        orientation(Integer): Orientation of the content considering rotation.
            See repository.RepositoryFile for acceptable values.
        creation_date(DateTime): Creation date of the file content.
        last_modified(DateTime): Date of last file modification.
        last_updated(DateTime): Date of last metadata update.
        description(String(255)): Description of the file.
        rating(Integer): Star rating of the file content.
        latitude(Float): Latitude of geographical coordinates
        longitude(Float): Longitude of geographical coordinates
        altitude(Float): Altitude of geopgraphical coordinates
        random_number(Float): Random number between 0..1. Used for creating
          starting points of smart order index iterations.
        verified(Boolean): Verification flag set during index building. True if
            file exists. False if not yet verified or does not exist.
    """

    __tablename__ = "files"
    id = Column(Integer, primary_key=True)
    rep_uuid = Column(String(Repository.MAX_LEN_UUID), nullable=False)
    # File uuid needs to be unique within a repository, but not across repositories.
#    file_uuid = Column(String(255), unique=True, nullable=False)
    file_uuid = Column(String(255), nullable=False)
    name = Column(String(255), nullable=False, index=True)
    type = Column(Integer)
    width = Column(Integer)
    height = Column(Integer)
    rotation = Column(Integer)
    orientation = Column(Integer)
    creation_date = Column(DateTime, index=True)
    description = Column(String(255))
    rating = Column(Integer)
    latitude = Column(Float)
    longitude = Column(Float)
    altitude = Column(Float)
    random_number = Column(Float, index=True)
    last_modified = Column(DateTime)
    last_updated = Column(DateTime)
    verified = Column(Boolean)
    tags = relationship("MetaDataTag", secondary="tag_file", backref=backref("files", lazy="dynamic"))


class MetaDataTag(Base):
    """Database model for tags in file metadata.

    Database representation of tags in file metadata using SQLAlchemy object
    relational mapper. Tags and metadata are connected via a separate table
    'tag_file' in the database.

    Properties:
        id(Integer): Numerical unique identifier. Automatically generated.
        name(String(255)): Unique name of the tag.
    """

    __tablename__ = "tags"
    id = Column(Integer, primary_key=True)
    name = Column(String(255), unique=True, nullable=False)


class Link(Base):
    """Association table for file metadata and tags."""

    __tablename__ = "tag_file"
    tag_id = Column(Integer, ForeignKey('tags.id', ondelete="CASCADE"), primary_key=True)
    file_id = Column(Integer, ForeignKey('files.id', ondelete="CASCADE"), primary_key=True)


class SORT_DIR(str, Enum):
    """Enumeration of index sort directions."""
    ASC = "ascending"
    DESC = "descending"


class SORT_ORDER(str, Enum):
    """Enumeration of index sort orders."""
    DATE = "date"
    NAME = "name"
    RANDOM = "random"
    SMART = "smart"


class Index:
    """File metadata index.

    Enables quick lookup and filtering of files based on their metadata. Index
    may span multiple repositories. Files and repositories are referenced via
    there universal unique identifiers.

    Use method build() to build metadata index. Method build() may be run in
    the background from a different thread.
    """

    # Required and valid index filter and sort criteria
    CRIT_REQ_KEYS = set()
    CRIT_VALID_KEYS = {'direction', 'excluded_tags', 'most_recent', 'order', 'orientation', 'repositories', 'smart_limit', 'smart_time', 'tags', 'types'} | CRIT_REQ_KEYS

    def __init__(self, dbname="index.sqlite"):
        """Initialize file index.

        :param dbname: Name of database. Default is "index.sqlite".
        :type dbname: str
        """
        self._dbname = dbname
        try:
            logging.info(f"Opening file index database '{dbname}'")
            # Determine whether we want verbose SQL debugging information.
            echo_flag = False
            if logging.getLogger("sqlalchemy").getEffectiveLevel() <= logging.DEBUG:
                echo_flag = True
            # Create sqlite database engine
            self._engine = create_engine(f"sqlite:///{dbname}", echo=echo_flag)
            # Create base class metadata
            Base.metadata.create_all(self._engine)
            # Open database session
            self._session_factory = sessionmaker(bind=self._engine)
            self._scoped_session = scoped_session(self._session_factory)
            self._session = self._scoped_session()
        except Exception as e:
            logging.critical(f"An error ocurred while opening the index database '{dbname}': {e}")

    def __del__(self):
        self.close()

    def build(self, rep, rebuild=False):
        """Build metadata index.

        Build method may be called from a different thread to build the index in
        the background. The method thus creates its own session and prevents
        files from looking up metadata from the index.

        :param rep: Repository for which to build the index.
        :type rep: repository.Repository
        :param rebuild: Indicates whether index is to be completely rebuilt.
            Default (False) is to update only, i.e. add the missing entries.
        :type rebuild: bool
        """
        # Create new session since build may be called from different thread.
        session = self._scoped_session()

        # Delete all metadata for the specified repository.
        if rebuild:
            logging.info(f"Rebuilding metadata index for repository '{rep.uuid}'.")
            try:
                logging.debug(f"Deleting all metadata entries of repository '{rep.uuid}'.")
                # Delete all file entries for the specified repository.
                session.query(MetaData).filter(MetaData.rep_uuid == rep.uuid).delete()
                session.commit()
                # Delete all unused tags.
                tags = session.query(MetaDataTag).all()
                for tag in tags:
                    if tag.files.count() == 0:
                        logging.info(f"Deleting unused tag '{tag.name}'.")
                        session.delete(tag)
                session.commit()
            except Exception as e:
                logging.error(f"An error occurred while deleting metadata of repository {rep.uuid} from index: {e}")
        # Mark existing metadata entries for verification.
        else:
            logging.info(f"Updating metadata index for repository '{rep.uuid}'.")
            try:
                logging.debug(f"Resetting verification flags for all metadata entries of repository '{rep.uuid}'.")
                query = update(MetaData).where(MetaData.rep_uuid == rep.uuid).values(verified=False)
                session.execute(query)
                session.commit()
            except Exception as e:
                logging.error(f"An error occurred while marking metadata entries of repository '{rep.uuid}' for verification: {e}")

        # Iterate through all files in the repository.
        for file in rep.iterator(index_lookup=False, extract_metadata=False):
            try:
                # Create new metadata entry for file if not included in the
                # index yet or outdated.
                mdata = session.query(MetaData).filter(MetaData.rep_uuid == rep.uuid).filter(MetaData.file_uuid == file.uuid).first()
                if mdata is None or mdata.last_updated < file.last_modified:
                    # Extract metadata from file.
                    file.extract_metadata()

                    # Create all necessary tags in database.
                    tags = list()
                    if file.tags:
                        for name in file.tags:
                            # Try to query tag from database.
                            tag = session.query(MetaDataTag).filter(MetaDataTag.name == name).first()
                            # Create and add tag to database otherwise.
                            if tag is None:
                                logging.info(f"Adding tag '{name}'.")
                                tag = MetaDataTag(name=name)
                            tags.append(tag)

                    # Create/update metadata entry with file metadata.
                    if mdata is None:
                        logging.info(f"Adding metadata of file '{file.uuid}' to index.")
                    else:
                        logging.info(f"Updating metadata of file '{file.uuid}' in index.")
                    mdata = MetaData(rep_uuid=file.rep.uuid,
                        file_uuid=file.uuid,
                        name=file.name,
                        type=file.type,
                        width=file.width,
                        height=file.height,
                        rotation=file.rotation,
                        orientation=file.orientation,
                        creation_date=file.creation_date,
                        last_modified=file.last_modified,
                        last_updated=file.last_updated,
                        description=file.description, rating=file.rating,
                        latitude=file._coordinates[0],
                        longitude=file._coordinates[1],
                        altitude=file._coordinates[2],
                        random_number=random.random(),
                        verified=True,
                        tags=tags)
                    session.add(mdata)
                    # Commit all changes to the database.
                    session.commit()
                else:
                    logging.debug(f"Skipping file '{file.uuid} as already included in index.")
                    # Mark entry as verified.
                    query = update(MetaData).where(MetaData.rep_uuid == rep.uuid).where(MetaData.file_uuid == file.uuid).values(verified=True)
                    session.execute(query)
                    # Do not immediately commit update for performance reasons.
#                    session.commit()
            except Exception as e:
                logging.error(f"An error occurred while building the metadata index: {e}")

        # Delete entries which have not been successfulyy verified.
        query = delete(MetaData).where(MetaData.verified == False)
        session.execute(query)
        # Commit pending changes and close session
        session.commit()
        session.close()

    def close(self):
        try:

            # Close database session and dispose engine
            if self._session:
                logging.debug(f"Closing database session.")
                self._session.close()
                self._session = None
            if self._engine:
                logging.debug(f"Disposing engine for database '{self._dbname}'.")
                self._engine.dispose()
                self._engine = None
        except Exception as e:
            logging.error(f"An error ocurred while closing the index database '{self._dbname}': {e}")

    def lookup(self, file, rep):
        """Lookup file metadata.

        :param file: File for which to lookup metadata.
        :type file: repository.Index
        :param rep: Repository containing the file.
        :type rep: repository.Repository
        :return: Returns a metadata object for the file if available. Returns
            None otherwise.
        :return type: repository.MetaData
        """
        try:
            mdata = self._session.query(MetaData).filter(MetaData.rep_uuid == rep.uuid).filter(MetaData.file_uuid == file.uuid).first()
            return mdata
        except Exception as e:
            logging.error(f"An error ocurred while looking up metadata from index for file '{file.uuid}' in repository '{rep.uuid}': {e}")
            return None

    def count(self):
        """Count the number of rows in the index.

        :return: Number of rows in the index.
        :return type: int
        """
        return self._session.query(MetaData).count()

    def iterator(self, **criteria):
        """Return selective iterator.

        Return selective iterator which allows to traverse through a
        sub-population of files in the index according to specified criteria.

        :param criteria: Selection criteria
        :type criteria: dict
        :return: Selective iterator
        :return type: repository.IndexIterator
        """
        return IndexIterator(self._session, **criteria)


class IndexIterator:
    """Selective index iterator.

    Selective iterator which allows to traverse through a sub-population of
    files in the index according to specified filter and order criteria.
    """

    def __init__(self, session, **criteria):
        """Initialize selective index iterator.

        :param session: SQLAlchemy database session
        :type session: sqlalchemy.orm.Session
        :param criteria: dictionary containing iteration criteria
        :type criteria: dict
        :raises: ConfigError
        """
        self._criteria = criteria
        self._result = None
        self._length = 0
        self._position = 0

        # Check the configuration for valid and required parameters.
        check_valid_required(criteria, Index.CRIT_VALID_KEYS, Index.CRIT_REQ_KEYS)
        # Check parameters.
        check_param('direction', criteria, required=False, options={ item.value for item in SORT_DIR })
        check_param('excluded_tags', criteria, required=False, recurse=True, is_str=True)
        check_param('most_recent', criteria, required=False, gr=0)
        check_param('order', criteria, required=False, options={ item.value for item in SORT_ORDER })
        check_param('orientation', criteria, required=False, options={ RepositoryFile.ORIENTATION_PORTRAIT, RepositoryFile.ORIENTATION_LANDSCAPE })
        check_param('repositories', criteria, required=False, recurse=True, is_str=True)
        check_param('tags', criteria, required=False, recurse=True, is_str=True)
        check_param('types', criteria, required=False, options={ RepositoryFile.TYPE_IMAGE, RepositoryFile.TYPE_VIDEO })
        # Check smart order related parameters.
        if criteria.get('orientation') == SORT_ORDER.SMART:
            check_param('smart_limit', criteria, required=True, is_int=True, gr=0)
            check_param('smart_time', criteria, required=True, gr=0)

        # Initialize query.
        query = session.query(MetaData.file_uuid, MetaData.rep_uuid, MetaData.creation_date)

        # Extend query based on iteration criteria.
        for key, value in criteria.items():

            # Filter for repository by UUID
            if key == "repositories":
                if isinstance(value, (list, set)):
                    query = query.filter(MetaData.rep_uuid.in_(value))
                else:
                    query = query.filter(MetaData.rep_uuid == value)

            # Filter for file type.
            elif key == "types":
                if isinstance(value, (list, set)):
                    query = query.filter(MetaData.type.in_(value))
                else:
                    query = query.filter(MetaData.type == value)

            # Filter for orientation of content.
            elif key == "orientation":
                query = query.filter(MetaData.orientation == value)

            # Limit iteration to files with specified tags.
            elif key == "tags":
                # Convert to list if single value specified.
                if type(value) == str: value = [value]
                query = query.filter(MetaData.tags.any(func.lower(MetaDataTag.name).in_([tag.lower() for tag in value])))

            # Exclude files with excluded tags from iteration.
            elif key == "excluded_tags":
                # Convert to list if single value specified.
                if type(value) == str: value = [value]
                query = query.filter(or_(~MetaData.tags.any(func.lower(MetaDataTag.name).in_([tag.lower() for tag in value])), MetaData.tags == None))

        # Determine limiting date to limit iteration to the n most recent files
        # based on the creation date. Outside of the loop since query needs to
        # be executed after all filters, but prior to specifying any order
        # (see below).
        if 'most_recent' in criteria:
            value = criteria['most_recent']
            result = query.order_by(desc(MetaData.creation_date)).limit(value).all()
            if len(result) > 0:
                date_limit = result[-1].creation_date
                query = query.filter(MetaData.creation_date >= date_limit)

        # Retrieve files in a specific or random order.
        if 'order' in criteria:
            order = criteria['order']
            dir = criteria.get('direction', SORT_DIR.ASC)
            map = { SORT_DIR.ASC: asc, SORT_DIR.DESC: desc }
            dir_fun = map[dir]
            if order == SORT_ORDER.RANDOM:
                query = query.order_by(func.random())
            elif order == SORT_ORDER.DATE:
                query = query.order_by(dir_fun(MetaData.creation_date))
            elif order == SORT_ORDER.NAME:
                query = query.order_by(dir_fun(func.upper(MetaData.name)))
            elif order == SORT_ORDER.SMART:
                logging.debug("Determine start date for smart order iteration.")
                # Metadata entries have been assigned a random number in the
                # range from 0..1 during creation. We now generate another
                # random number in the same range and retrieve the metadata
                # entry with the closest match as starting point for the series.
                random_number = random.random()
                result = query.filter(MetaData.random_number >= random_number).order_by(MetaData.random_number).first()
                if result is None:
                    result = query.filter(MetaData.random_number <= random_number).order_by(desc(MetaData.random_number)).first()
                if result is not None:
                    query = query.order_by(MetaData.creation_date).filter(MetaData.creation_date >= result.creation_date).limit(criteria['smart_limit'])

        # Query data and save list of metadata objects.
        logging.debug("Querying files for new iteration.")
        self._result = query.all()
        logging.debug(f"New iteration has {len(self._result)} files.")

    def __iter__(self):
        """Return self as iterator.

        :return: self
        :return type: repository.IndexIterator
        """
        return self

    def __next__(self):
        """Return next file in iteration.

        :return: next file in iteration
        :return type: repository.RepositoryFile
        :raises: StopIteration
        """
        # Repeat until we have a valid file or end of iteration is reached.
        while True:
            # Retrieve next metadata object in iteration.
            if self._position < len(self._result):
                mdata = self._result[self._position]
                self._position = self._position + 1
            # Raise exception if end of iteration is reached.
            else:
                raise StopIteration()
            # Evaluate smart order termination criteria from second file onwards.
            if self._criteria.get('order') == SORT_ORDER.SMART and self._position > 1:
                # Obtain meta data of previous file.
                prev_mdata = self._result[self._position - 2]
                # Stop iteration if delta in creation date between current and
                # previous file exceeds smart time limit.
                delta = mdata.creation_date - prev_mdata.creation_date
                delta = delta.seconds/3600 + delta.days*24
                if delta > self._criteria['smart_time']:
                    logging.debug("Ending iteration early due to smart time criterion.")
                    raise StopIteration()
                # We additionally plan to implement a smart distance criterion
                # at a later point in time once location meta data are supported.
            # Try to obtain corresponding file.
            try:
                return Repository.by_uuid(mdata.rep_uuid).file_by_uuid(mdata.file_uuid)
            # Catch any invalid uuid errors in case the file is no longer
            # available in the repository and continue.
            except UuidError:
                logging.warn(f"Skipping invalid file '{mdata.file_uuid}' in repository '{mdata.rep_uuid}'. The metadata index may be outdated.")
                pass

    @property
    def length(self):
        """Return number of files in index."""
        return len(self._result)

    def previous(self, n=1):
        """Return previous file in iteration.

        Optionally, an offset may be specified to retrieve even earlier elements
        of the iteration.

        :param n: Optional offset. An offset of 1 corresponds to the previous
          element. An offset of 2 to the second previous element etc.
        :type n: int
        :return: previous file in iteration
        :return type: repository.RepositoryFile
        :raises: StopIteration
        """
        # Adjust position to retrieve previous as next file.
        if self._position > n:
            self._position = self._position - (n + 1)
            return next(self)
        else:
            raise StopIteration()



"""Module providing slideshow image class using PIL for image loading and manipulation."""

import random

from kivy.animation import Animation
from kivy.graphics import Color, Rectangle
from kivy.graphics.texture import Texture
from kivy.graphics.transformation import Matrix
from kivy.logger import Logger
from kivy.uix.image import Image
from kivy.uix.scatter import Scatter

from PIL import Image as PilImage

from .base import LabeledContent


class SlideshowImage(LabeledContent):
    """Image slideshow widget.

    Loads the image from the specified File and starts playing it as soon as the
    widget becomes visible. The image is scaled to fit the entire widget,
    respecting the aspect ratio.
    """

    def __init__(self, file, config):
        """Initialize slideshow image instance.

        :param file: Repository file instance for the image to be displayed.
        :type file: repository.File
        :param config: Dictionary with the following entries:
            rotation: Angle in degrees (int) by which the image is rotated clockwise.
            bgolor: Canvas background color (list(3)) for areas, which are not covered by the image.
            resize: Mode (str) for resizing of images. Must equal "fit" or "fill".
        :type config: dict
        """
        super().__init__(file, config)
        # Calculate total rotation from image rotation and rotation configured for the slideshow.
        # To avoid negative values and values beyond 360°, we additionally apply a modulo operation.
        self._rotation = (config['rotation'] - file.rotation) % 360
        self._bgcolor = config['bg_color']
        self._resize = config['resize']
        self._pause = config['pause']
        # Create nested image and scatter widgets.
        self._image = Image()
        self._scatter = Scatter()
        self._scatter.add_widget(self._image)
        self.add_widget(self._scatter, len(self.children))

        if(config['enable_animation']):
            # Initial values for zoom animation.
            self._zoom = 1.0
            self._anchor = (0, 0)
            # Create zoom animation.            
            self._animation = Animation(_zoom=config['max_zoom'], duration=config['pause'], t='in_out_sine')
            self._animation.bind(on_progress=self.on_progress)
            self._animation.start(self)
        
        # Call update_canvas method when the size of the widget changes.
        self.bind(size=self.update_canvas)


    def on_progress(self, *args):
        """Adjust zoom level for the next animation frame."""
        # Reset transformation matrix
        self._scatter.transform = Matrix().identity()
        # Apply new zoom factor
        matrix = Matrix().scale(self._zoom, self._zoom, self._zoom)
        self._scatter.apply_transform(matrix, anchor=self._anchor)        


    def update_canvas(self, *args):
        """Update canvas when the size of the widget changes."""

        # Adjust image size to parent widget size.
        self._image.size = self.size

        # Clear before and after groups of image canvas.
        self._image.canvas.before.clear()
        self._image.canvas.after.clear()

        # Fill image canvas with background color.
        with self._image.canvas.before:
            Color(*self._bgcolor)
            Rectangle(pos=(0, 0), size=self.size)

        # Load image with PIL.
        pil_image = PilImage.open(self._file.source)
        #pil_image = pil_image.transpose(PilImage.Transpose.FLIP_TOP_BOTTOM)

        # Rotate image as required.
        if self._rotation == 90:
            pil_image = pil_image.transpose(PilImage.Transpose.ROTATE_90)
        elif self._rotation == 180:
            pil_image = pil_image.transpose(PilImage.Transpose.ROTATE_180)
        elif self._rotation == 270:
            pil_image = pil_image.transpose(PilImage.Transpose.ROTATE_270)
        
        # Determine aspect ratios of image slideshow widget (this widget)
        # and image.
        widget_ratio = self.width/self.height
        image_ratio = pil_image.width/pil_image.height 

        # Tranform image to fill the image slideshow widget. Only images with
        # the same orientation will be resized to fill the widget. Images with
        # a different orientation will be resized to fit the widget.
        if self._resize == "fill":
            if widget_ratio >= image_ratio and image_ratio >= 1:
                self._image.fit_mode = "cover"
                width = self.width
                height = int(self.width/image_ratio)
            elif widget_ratio >= image_ratio and image_ratio < 1:
                self._image.fit_mode = "contain"
                width = int(image_ratio * self.height)
                height = self.height
            elif widget_ratio < image_ratio and image_ratio >= 1:
                self._image.fit_mode = "cover"
                width = int(image_ratio * self.height)
                height = self.height
            else:
                self._image.fit_mode = "contain"
                width = self.width
                height = int(self.width/image_ratio)
        
        # Default is to fit the image to the canvas
        else:  # self._resize == "fit"
            self._image.fit_mode = "contain"
            if widget_ratio >= image_ratio:
                width = int(image_ratio * self.height)
                height = self.height
            else:
                width = self.width
                height = int(self.width/image_ratio)

        # Resize PIL image.
        pil_image = pil_image.resize((width, height), PilImage.Resampling.LANCZOS)

        # Create texture from PIL image and assign to Kivy image widget.
        texture = Texture.create(size=(pil_image.width, pil_image.height), mipmap=False)
        texture.blit_buffer(pil_image.tobytes(), colorfmt='rgb', bufferfmt='ubyte')
        texture.flip_vertical()
        self._image.texture = texture

        # Set anchor for the zoom animation.
        if self._config['enable_animation']:
            # Widget and image have same orientation.
            if (widget_ratio >= 1 and image_ratio >= 1) or (widget_ratio < 1 and image_ratio < 1):
                self._anchor = (random.randint(0, self.width), random.randint(int(self.height/2), self.height))
            # Widget has landscape and image portrait orientation.
            elif widget_ratio >= 1 and image_ratio < 1:
                self._anchor = (int(self.width/2), random.randint(int(self.height/2), self.height))
            # Widget has portrait and image landscape orientation.
            elif widget_ratio < 1 and image_ratio >= 1:
                self._anchor = (random.randint(0, self.width), int(self.height/2))


        # Log debug information
#        Logger.debug(f"Image uuid: {self._file.uuid}")
#        Logger.debug(f"Image type: {self._file.type}")
#        Logger.debug(f"Image source: {self._file.source}")
#        Logger.debug(f"Image orientation: {self._file.orientation}")
#        Logger.debug(f"Image rotation: {self._file.rotation}")
#        Logger.debug(f"Total rotation: {self._rotation}")
#        Logger.debug(f"Widget width: {self.width}")
#        Logger.debug(f"Widget height: {self.height}")
#        Logger.debug(f"Widget aspect ratio: {widget_ratio}")
#        Logger.debug(f"max_dim: {max_dim}")
#        Logger.debug(f"Image width: {self._image.width}")
#        Logger.debug(f"Image height: {self._image.height}")
#        Logger.debug(f"Image aspect ratio: {image_ratio}")
#        Logger.debug(f"Image x: {self._image.x}")
#        Logger.debug(f"Image y: {self._image.y}")
#        Logger.debug(f"Image center: {self._image.center}")
